from os.path import exists, expanduser
from typing import List
from pprint import pformat

from pytezos.operation.result import OperationResult
from pytezos.michelson.contract import Contract
from pytezos.michelson.converter import convert
from pytezos.michelson.micheline import skip_nones
from pytezos.michelson.formatter import micheline_to_michelson
from pytezos.operation.group import OperationGroup
from pytezos.operation.content import format_mutez, format_tez
from pytezos.interop import Interop
from pytezos.tools.docstring import get_class_docstring
from pytezos.repl.interpreter import Interpreter


class ContractCallResult(OperationResult):
    """ Encapsulates the result of a contract invocation.
    """

    @classmethod
    def from_contract_call(cls, operation_group: dict, address, contract: Contract) -> list:
        """ Get a list of results from an operation group content with metadata.

        :param operation_group: {..., "contents": [{..., kind: "transaction", ...}]}
        :param address: address of the invoked contract
        :param contract: invoked contract
        :rtype: List[ContractCallResult]
        """
        results = list()
        for content in OperationResult.iter_contents(operation_group):
            if content['kind'] == 'transaction':
                if content['destination'] == address:
                    results.append(cls.from_transaction(content))
            elif content['kind'] == 'origination':
                result = cls.get_result(content)
                if address in result.get('originated_contracts', []):
                    results.append(cls.from_origination(content))

        def decode_result(res):
            kwargs = dict(storage=contract.storage.decode(res.storage))
            if hasattr(res, 'big_map_diff'):
                contract.storage.big_map_init(res.storage)
                kwargs.update(big_map_diff=contract.storage.big_map_diff_decode(res.big_map_diff))
            if hasattr(res, 'parameters'):
                kwargs.update(parameters=contract.parameter.decode(data=res.parameters))
            if hasattr(res, 'operations'):
                kwargs.update(operations=res.operations)
            return cls(**kwargs)

        return list(map(decode_result, results))

    @classmethod
    def from_code_run(cls, code_run: dict, parameters, contract: Contract):
        """ Parse a result of `run_code` execution.

        :param code_run: RPC response (json)
        :param parameters: Micheline expression
        :param contract: invoked contract
        :rtype: ContractCallResult
        """
        contract.storage.big_map_init(code_run['storage'])
        return cls(
            parameters=contract.parameter.decode(parameters),
            storage=contract.storage.decode(code_run['storage']),
            big_map_diff=contract.storage.big_map_diff_decode(code_run.get('big_map_diff', [])),
            operations=code_run.get('operations', [])
        )

    @classmethod
    def from_repl_result(cls, res: dict, parameters, contract: Contract):
        """ Parse an output of the builtin interpreter.

        :param res: Interpreter output
        :param parameters: Micheline expression
        :param contract: invoked contract
        :returns: ContractCallResult
        """
        contract.storage.big_map_init(res['result']['storage'].val_expr)
        return cls(
            parameters=contract.parameter.decode(parameters),
            storage=contract.storage.decode(res['result']['storage'].val_expr),
            big_map_diff=contract.storage.big_map_diff_decode(res['result']['big_map_diff']),
            operations=[x.content for x in res['result']['operations']]
        )


class ContractCall(Interop):
    """ Proxy class encapsulating a contract call: contract type scheme, contract address, parameters, and amount
    """

    def __init__(self, parameters,
                 address=None, contract: Contract = None, factory=Contract, amount=0, shell=None, key=None):
        super(ContractCall, self).__init__(shell=shell, key=key)
        self.parameters = parameters
        self.address = address
        self.amount = amount

        if contract is None:
            assert address is not None
            contract = factory.from_micheline(self.shell.contracts[address].code())

        self.contract = contract

    def _spawn(self, **kwargs):
        return ContractCall(
            parameters=self.parameters,
            address=self.address,
            contract=self.contract,
            amount=kwargs.get('amount', self.amount),
            shell=kwargs.get('shell', self.shell),
            key=kwargs.get('key', self.key)
        )

    def __repr__(self):
        res = [
            super(ContractCall, self).__repr__(),
            f'.address  # {self.address}',
            f'.amount  # {self.amount}',
            '\nParameters',
            pformat(self.parameters),
            '\nHelpers',
            get_class_docstring(self.__class__)
        ]
        return '\n'.join(res)

    def with_amount(self, amount):
        """ Send funds to the contract too.

        :param amount: amount in microtez (int) or tez (Decimal)
        :rtype: ContractCall
        """
        return self._spawn(amount=amount)

    @property
    def operation_group(self) -> OperationGroup:
        """ Show generated operation group.

        :rtype: OperationGroup
        """
        return OperationGroup(shell=self.shell, key=self.key) \
            .transaction(destination=self.address,
                         amount=self.amount,
                         parameters=self.parameters) \
            .fill()

    def inject(self):
        """ Autofill, sign and inject resulting operation group.
        """
        return self.operation_group.autofill().sign().inject()

    def cmdline(self):
        """ Generate command line for tezos client.
        """
        arg = micheline_to_michelson(self.parameters['value'], inline=True)
        source = self.key.public_key_hash()
        amount = format_tez(self.amount)
        entrypoint = self.parameters['entrypoint']
        return f'transfer {amount} from {source} to {self.address} ' \
               f'--entrypoint \'{entrypoint}\' --arg \'{arg}\''

    def interpret(self, storage, source=None, sender=None, amount=None, balance=None, chain_id=None, now=None):
        """ Run code in the builtin REPL (WARNING! Not recommended for critical tasks).

        :param storage: Python object
        :param source: patch SOURCE
        :param sender: patch SENDER
        :param amount: patch AMOUNT
        :param balance: patch BALANCE
        :param chain_id: patch CHAIN_ID
        :param now: patch NOW
        :rtype: ContractCallResult
        """
        i = Interpreter()
        i.execute(self.contract.text)

        if source is None:
            source = self.key.public_key_hash()
        if sender is None:
            sender = source
        if amount is None:
            amount = 0
        if balance is None:
            balance = 0

        patch_map = {
            'SOURCE': source,
            'SENDER': sender,
            'AMOUNT': amount,
            'BALANCE': balance,
            'CHAIN_ID': chain_id,
            'NOW': now
        }
        for instr, value in patch_map.items():
            if value is not None:
                value = f'"{value}"' if isinstance(value, str) else value
                i.execute(f'PATCH {instr} {value}')

        s_expr = micheline_to_michelson(self.contract.storage.encode(storage), inline=True, wrap=True)
        p_expr = micheline_to_michelson(self.parameters['value'], inline=True, wrap=True)
        res = i.execute(f'RUN %{self.parameters["entrypoint"]} {p_expr} {s_expr}')

        return ContractCallResult.from_repl_result(
            res, parameters=self.parameters, contract=self.contract)

    def result(self, storage=None, source=None, sender=None, gas_limit=None):
        """ Simulate operation and parse the result.

        :param storage: Python object only. If storage is specified, `run_code` is called instead of `run_operation`.
        :param source: Can be specified for unit testing purposes
        :param sender: Can be specified for unit testing purposes, \
        see https://tezos.gitlab.io/whitedoc/michelson.html#operations-on-contracts for the difference
        :param gas_limit: Specify gas limit (default is gas hard limit)
        :rtype: ContractCallResult
        """
        chain_id = self.shell.chains.main.chain_id()
        if storage is not None or source or sender or gas_limit:
            query = skip_nones(
                script=self.contract.code,
                storage=self.contract.storage.encode(storage),
                entrypoint=self.parameters['entrypoint'],
                input=self.parameters['value'],
                amount=format_mutez(self.amount),
                chain_id=chain_id,
                source=sender,
                payer=source,
                gas=str(gas_limit) if gas_limit is not None else None
            )
            code_run_res = self.shell.head.helpers.scripts.run_code.post(query)
            return ContractCallResult.from_code_run(
                code_run_res, parameters=self.parameters, contract=self.contract)
        else:
            opg_with_metadata = self.operation_group.fill().run()
            res = ContractCallResult.from_contract_call(
                opg_with_metadata, address=self.address, contract=self.contract)
            return res[0] if res else None

    def view(self):
        """ Get return value of a view method.

        :returns: object
        """
        opg_with_metadata = self.operation_group.fill().run()
        view_operation = OperationResult.get_contents(opg_with_metadata, source=self.address)[0]
        view_contract = Contract.from_micheline(self.shell.contracts[view_operation['destination']].code())
        return view_contract.parameter.decode(view_operation['parameters'])


class ContractEntrypoint(Interop):
    """ Proxy class for spawning ContractCall instances.
    """

    def __init__(self, name, address=None, contract: Contract = None, factory=Contract, shell=None, key=None):
        super(ContractEntrypoint, self).__init__(shell=shell, key=key)
        if contract is None:
            assert address is not None
            code = self.shell.contracts[address].code()
            contract = factory.from_micheline(code)

        self.contract = contract
        self.name = name
        self.address = address

    def _spawn(self, **kwargs):
        return ContractEntrypoint(
            name=self.name,
            contract=self.contract,
            address=self.address,
            shell=kwargs.get('shell', self.shell),
            key=kwargs.get('key', self.key),
        )

    def __repr__(self):
        res = [
            super(ContractEntrypoint, self).__repr__(),
            f'.address  # {self.address}',
            f'\n{self.__doc__}'
        ]
        return '\n'.join(res)

    def __call__(self, *args, **kwargs):
        """ Spawn a contract call proxy initialized with the entrypoint name

        :param args: entrypoint args
        :param kwargs: entrypoint key-value args
        :rtype: ContractCall
        """
        if args:
            if len(args) == 1:
                (data, is_single) = (args[0], True)
            else:
                (data, is_single) = (list(args), False)
        elif kwargs:
            (data, is_single) = (kwargs, False)
        else:
            (data, is_single) = ([], False)

        if self.name:
            data = {self.name: data} if data or is_single else self.name

        parameters = self.contract.parameter.encode(data)
        return ContractCall(
            parameters=parameters,
            address=self.address,
            contract=self.contract,
            shell=self.shell,
            key=self.key,
        )


class ContractInterface(Interop):
    """ Proxy class for interacting with a contract.
    """
    __default_entry__ = 'call'

    def __init__(self, address=None, contract: Contract = None, factory=Contract, shell=None, key=None):
        super(ContractInterface, self).__init__(shell=shell, key=key)
        if contract is None:
            assert address is not None
            code = self.shell.contracts[address].code()
            contract = factory.from_micheline(code)

        self.contract = contract
        self.address = address

        for entry_name, docstring in contract.parameter.entries(default=self.__default_entry__):
            entry_point = ContractEntrypoint(
                name=entry_name if entry_name != self.__default_entry__ else None,
                address=self.address,
                contract=contract,
                shell=self.shell,
                key=self.key
            )
            entry_point.__doc__ = docstring
            setattr(self, entry_name, entry_point)

    def _spawn(self, **kwargs):
        return ContractInterface(
            address=self.address,
            contract=self.contract,
            shell=kwargs.get('shell', self.shell),
            key=kwargs.get('key', self.key)
        )

    def __repr__(self):
        entrypoints, _ = zip(*self.contract.parameter.entries(default=self.__default_entry__))
        res = [
            super(ContractInterface, self).__repr__(),
            f'.address  # {self.address}',
            '\nEntrypoints',
            *list(map(lambda x: f'.{x}()', entrypoints)),
            '\nHelpers',
            get_class_docstring(self.__class__,
                                attr_filter=lambda x: not x.startswith('_') and x not in entrypoints)
        ]
        return '\n'.join(res)

    @classmethod
    def create_from(cls, source, shell=None, factory=Contract):
        """ Initialize from contract code.

        :param source: Michelson code
        :param shell: A `Shell` instance (optional)
        :param factory: An optional contract factory if you need to force interface (forced annotations)
        :rtype: ContractInterface
        """
        if isinstance(source, str) and exists(expanduser(source)):
            contract = factory.from_file(source)
        else:
            contract = factory(convert(source, output='micheline'))

        return ContractInterface(contract=contract, shell=shell)

    def big_map_get(self, path, block_id='head'):
        """ Get BigMap entry as Python object by plain key and block height.

        :param path: Json path to the key (or just key to access default BigMap location). \
            Use `/` to separate nodes and `:` to separate tuple args. \
            In any other case you'd need to escape those symbols.
        :param block_id: Block height / hash / offset to use, default is `head`
        :returns: object
        """
        if not self.contract.storage.big_map_schema:
            storage = self.shell.blocks[block_id].context.contracts[self.address].storage()
            self.contract.storage.big_map_init(storage)

        query = self.contract.storage.big_map_query(path)
        if query.get('big_map_id'):
            value = self.shell.blocks[block_id].context.big_maps[query['big_map_id']][query['script_expr']]()
        else:
            value = self.shell.blocks[block_id].context.contracts[self.address].big_map_get.post(query)
        return self.contract.storage.big_map_decode(value, query.get('big_map_id'))

    def storage(self, block_id='head'):
        """ Get storage as Pythons object at specified block height.

        :param block_id: Block height / hash / offset to use, default is `head`
        :returns: object
        """
        storage = self.shell.blocks[block_id].context.contracts[self.address].storage()
        return self.contract.storage.decode(storage)

    def operation_result(self, operation_group: dict) -> List[ContractCallResult]:
        """ Get operation parameters, storage and big_map_diff as Python objects.
        Can locate operation inside operation groups with multiple contents and/or internal operations.

        :param operation_group: {'branch', 'protocol', 'contents', 'signature'}
        :rtype: ContractCallResult
        """
        return ContractCallResult.from_contract_call(
            operation_group, address=self.address, contract=self.contract)

    def manager(self):
        """ Get contract manager address (tz).
        """
        return self.shell.block.context.contracts[self.address].manager()



from os.path import exists, expanduser

from pytezos.rpc import ShellQuery, RpcNode, mainnet, babylonnet, carthagenet, zeronet, localhost, labnet, pool
from pytezos.crypto import Key, is_installed
from pytezos.encoding import is_key, is_pkh
from pytezos.tools.docstring import InlineDocstring

default_shell = 'carthagenet'
default_key = 'edsk33N474hxzA4sKeWVM6iuGNGDpX2mGwHNxEA4UbWS8sW3Ta3NKH'  # please, use responsibly
default_key_hash = 'tz1cnQZXoznhduu4MVWfJF6GSyP6mMHMbbWa'

alice_key = 'edsk3QoqBuvdamxouPhin7swCvkQNgq4jP5KZPbwWNnwdZpSpJiEbq'
alice_key_hash = 'tz1VSUr8wwNhLAzempoch5d6hLRiTh8Cjcjb'


class KeyHash(Key):

    def __init__(self, public_key_hash):
        super(KeyHash, self).__init__(0)
        self._pkh = public_key_hash

    def __repr__(self):
        res = [
            super(Key, self).__repr__(),
            f'\nPublic key hash',
            self.public_key_hash()
        ]
        return '\n'.join(res)

    def public_key_hash(self):
        return self._pkh

    def public_key(self):
        raise NotImplementedError

    def secret_key(self, passphrase=None, ed25519_seed=True):
        raise NotImplementedError

    def sign(self, message, generic=False):
        raise NotImplementedError

    def verify(self, signature, message):
        raise NotImplementedError


class Interop(metaclass=InlineDocstring):
    """ Mixin for blockchain interaction, stores node connection and key object.
    """

    def __repr__(self):
        res = [
            super(Interop, self).__repr__(),
            '\nProperties',
            f'.key  # {self.key.public_key_hash()}',
            f'.shell  # {self.shell.node.uri} ({self.shell.node.network})'
        ]
        return '\n'.join(res)

    def __init__(self, shell=None, key=None):
        if shell is None:
            shell = default_shell

        if isinstance(shell, str):
            networks = {
                'mainnet': mainnet,
                'babylonnet': babylonnet,
                'carthagenet': carthagenet,
                'zeronet': zeronet,
                'sandboxnet': localhost.sandboxnet,
                'bbbox': localhost.bbbox,
                'labnet': labnet,
                'mainnet-pool': pool.mainnet
            }
            if shell in networks:
                self.shell = networks[shell]
            else:
                self.shell = ShellQuery(node=RpcNode(uri=shell))
        elif isinstance(shell, ShellQuery):
            self.shell = shell
        else:
            raise NotImplementedError(shell)

        if key is None:
            key = default_key if is_installed() else default_key_hash

        if isinstance(key, str):
            keys = {
                'alice': alice_key
            }
            if key in keys:
                self.key = keys[key]
            elif is_key(key):
                self.key = Key.from_encoded_key(key)
            elif is_pkh(key):
                self.key = KeyHash(key)
            elif exists(expanduser(key)):
                self.key = Key.from_faucet(key)
            else:
                self.key = Key.from_alias(key)
        elif isinstance(key, Key):
            self.key = key
        else:
            raise NotImplementedError(key)

    def _spawn(self, **kwargs):
        raise NotImplementedError

    def using(self, shell: ShellQuery = None, key: Key = None):
        """ Change current rpc endpoint and account (private key).

        :param shell: one of 'babylonnet', 'mainnet', 'zeronet', or RPC node uri, or instance of `ShellQuery`
        :param key: base58 encoded key, path to the faucet file, alias from tezos-client, or instance of `Key`
        :returns: A copy of current object with changes applied
        """
        return self._spawn(shell=shell, key=key)



import base58
import calendar
from datetime import datetime


def tb(l):
    return b''.join(map(lambda x: x.to_bytes(1, 'big'), l))


base58_encodings = [
    #    Encoded   |               Decoded             |
    # prefix | len | prefix                      | len | Data type
    (b"B",     51,   tb([1, 52]),                  32,   u"block hash"),
    (b"o",     51,   tb([5, 116]),                 32,   u"operation hash"),
    (b"Lo",    52,   tb([133, 233]),               32,   u"operation list hash"),
    (b"LLo",   53,   tb([29, 159, 109]),           32,   u"operation list list hash"),
    (b"P",     51,   tb([2, 170]),                 32,   u"protocol hash"),
    (b"Co",    52,   tb([79, 199]),                32,   u"context hash"),

    (b"tz1",   36,   tb([6, 161, 159]),            20,   u"ed25519 public key hash"),
    (b"tz2",   36,   tb([6, 161, 161]),            20,   u"secp256k1 public key hash"),
    (b"tz3",   36,   tb([6, 161, 164]),            20,   u"p256 public key hash"),
    (b"KT1",   36,   tb([2, 90, 121]),             20,   u"Originated address"),

    (b"id",    30,   tb([153, 103]),               16,   u"cryptobox public key hash"),

    (b'expr',  54,   tb([13, 44, 64, 27]),         32,   u'script expression'),
    (b"edsk",  54,   tb([13, 15, 58, 7]),          32,   u"ed25519 seed"),
    (b"edpk",  54,   tb([13, 15, 37, 217]),        32,   u"ed25519 public key"),
    (b"spsk",  54,   tb([17, 162, 224, 201]),      32,   u"secp256k1 secret key"),
    (b"p2sk",  54,   tb([16, 81, 238, 189]),       32,   u"p256 secret key"),

    (b"edesk", 88,   tb([7, 90, 60, 179, 41]),     56,   u"ed25519 encrypted seed"),
    (b"spesk", 88,   tb([9, 237, 241, 174, 150]),  56,   u"secp256k1 encrypted secret key"),
    (b"p2esk", 88,   tb([9, 48, 57, 115, 171]),    56,   u"p256_encrypted_secret_key"),

    (b"sppk",  55,   tb([3, 254, 226, 86]),        33,   u"secp256k1 public key"),
    (b"p2pk",  55,   tb([3, 178, 139, 127]),       33,   u"p256 public key"),
    (b"SSp",   53,   tb([38, 248, 136]),           33,   u"secp256k1 scalar"),
    (b"GSp",   53,   tb([5, 92, 0]),               33,   u"secp256k1 element"),

    (b"edsk",  98,   tb([43, 246, 78, 7]),         64,   u"ed25519 secret key"),
    (b"edsig", 99,   tb([9, 245, 205, 134, 18]),   64,   u"ed25519 signature"),
    (b"spsig", 99,   tb([13, 115, 101, 19, 63]),   64,   u"secp256k1 signature"),
    (b"p2sig", 98,   tb([54, 240, 44, 52]),        64,   u"p256 signature"),
    (b"sig",   96,   tb([4, 130, 43]),             64,   u"generic signature"),

    (b'Net',   15,   tb([87, 82, 0]),              4,    u"chain id"),
]

operation_tags = {
    'endorsement': 0,
    'seed_nonce_revelation': 1,
    'double_endorsement_evidence': 2,
    'double_baking_evidence': 3,
    'account_activation': 4,
    'proposal': 5,
    'ballot': 6,
    'reveal': 7,
    'transaction': 8,
    'origination': 9,
    'delegation': 10
}


def scrub_input(v) -> bytes:
    if isinstance(v, str) and not isinstance(v, bytes):
        try:
            _ = int(v, 16)
        except ValueError:
            v = v.encode('ascii')
        else:
            if v.startswith('0x'):
                v = v[2:]
            v = bytes.fromhex(v)

    if not isinstance(v, bytes):
        raise TypeError(
            "a bytes-like object is required (also str), not '%s'" %
            type(v).__name__)

    return v


def base58_decode(v: bytes) -> bytes:
    """ Decode data using Base58 with checksum + validate binary prefix against known kinds and cut in the end.

    :param v: Array of bytes (use string.encode())
    :returns: bytes
    """
    try:
        prefix_len = next(
            len(encoding[2])
            for encoding in base58_encodings
            if len(v) == encoding[1] and v.startswith(encoding[0])
        )
    except StopIteration:
        raise ValueError('Invalid encoding, prefix or length mismatch.')

    return base58.b58decode_check(v)[prefix_len:]


def base58_encode(v: bytes, prefix: bytes) -> bytes:
    """ Encode data using Base58 with checksum and add an according binary prefix in the end.

    :param v: Array of bytes
    :param prefix: Human-readable prefix (use b'') e.g. b'tz', b'KT', etc
    :returns: bytes (use string.decode())
    """
    try:
        encoding = next(
            encoding
            for encoding in base58_encodings
            if len(v) == encoding[3] and prefix == encoding[0]
        )
    except StopIteration:
        raise ValueError('Invalid encoding, prefix or length mismatch.')

    return base58.b58encode_check(encoding[2] + v)


def _validate(v, prefixes: list):
    v = scrub_input(v)
    if any(map(lambda x: v.startswith(x), prefixes)):
        base58_decode(v)
    else:
        raise ValueError('Unknown prefix.')


def validate_pkh(v):
    """ Ensure parameter is a public key hash (starts with b'tz1', b'tz2', b'tz3')

    :param v: string or bytes
    :raises ValueError: if parameter is not a public key hash
    """
    return _validate(v, prefixes=[b'tz1', b'tz2', b'tz3'])


def validate_sig(v):
    """ Ensure parameter is a signature (starts with b'edsig', b'spsig', b'p2sig', b'sig')

    :param v: string or bytes
    :raises ValueError: if parameter is not a signature
    """
    return _validate(v, prefixes=[b'edsig', b'spsig', b'p2sig', b'sig'])


def is_pkh(v) -> bool:
    """ Check if value is a public key hash.
    """
    try:
        validate_pkh(v)
    except (ValueError, TypeError):
        return False
    return True


def is_sig(v) -> bool:
    """ Check if value is a signature.
    """
    try:
        validate_sig(v)
    except (ValueError, TypeError):
        return False
    return True


def is_bh(v) -> bool:
    """ Check if value is a block hash.
    """
    try:
        _validate(v, prefixes=[b'B'])
    except (ValueError, TypeError):
        return False
    return True


def is_ogh(v) -> bool:
    """ Check if value is an operation group hash.
    """
    try:
        _validate(v, prefixes=[b'o'])
    except (ValueError, TypeError):
        return False
    return True


def is_kt(v) -> bool:
    """ Check if value is a KT address.
    """
    try:
        _validate(v, prefixes=[b'KT1'])
    except (ValueError, TypeError):
        return False
    return True


def is_key(v) -> bool:
    """ Check if value is a public key.
    """
    try:
        _validate(v, prefixes=[b"edsk", b"edpk", b"spsk", b"p2sk", b"sppk", b"p2pk"])
    except (ValueError, TypeError):
        return False
    return True


def is_chain_id(v) -> bool:
    """ Check if value is a chain id.
    """
    try:
        _validate(v, prefixes=[b'Net'])
    except (ValueError, TypeError):
        return False
    return True


def forge_nat(value) -> bytes:
    """ Encode a number using LEB128 encoding (Zarith).

    :param int value: the value to encode
    :returns: encoded value
    :rtype: bytes
    """
    if value < 0:
        raise ValueError('Value cannot be negative.')

    buf = bytearray()
    more = True

    while more:
        byte = value & 0x7f
        value >>= 7

        if value:
            byte |= 0x80
        else:
            more = False

        buf.append(byte)

    return bytes(buf)


def forge_public_key(value) -> bytes:
    """ Encode public key into bytes.

    :param value: public key in in base58 form
    """
    prefix = value[:4]
    res = base58.b58decode_check(value)[4:]

    if prefix == 'edpk':
        return b'\x00' + res
    elif prefix == 'sppk':
        return b'\x01' + res
    elif prefix == 'p2pk':
        return b'\x02' + res

    raise ValueError(f'Unrecognized key type: #{prefix}')


def parse_public_key(data: bytes) -> str:
    """ Decode public key from byte form.

    :param data: encoded public key.
    :returns: base58 encoded public key
    """
    key_prefix = {
        b'\x00': b'edpk',
        b'\x01': b'sppk',
        b'\x02': b'p2pk'
    }
    return base58_encode(data[1:], key_prefix[data[:1]]).decode()


def parse_chain_id(data: bytes):
    """ Decode chain id from byte form.

    :param data: encoded chain id.
    :returns: base58 encoded chain id
    """
    return base58_encode(data, b'Net').decode()


def parse_signature(data: bytes):
    """ Decode signature from byte form.

    :param data: encoded signature.
    :returns: base58 encoded signature (generic)
    """
    return base58_encode(data, b'sig').decode()


def forge_address(value: str, tz_only=False) -> bytes:
    """ Encode address or key hash into bytes.

    :param value: base58 encoded address or key_hash
    :param tz_only: True indicates that it's a key_hash (will be encoded in a more compact form)
    """
    prefix = value[:3]
    address = base58.b58decode_check(value)[3:]

    if prefix == 'tz1':
        res = b'\x00\x00' + address
    elif prefix == 'tz2':
        res = b'\x00\x01' + address
    elif prefix == 'tz3':
        res = b'\x00\x02' + address
    elif prefix == 'KT1':
        res = b'\x01' + address + b'\x00'
    else:
        raise ValueError(value)

    return res[1:] if tz_only else res


def parse_address(data: bytes):
    """ Decode address or key_hash from bytes.

    :param data: encoded address or key_hash
    :returns: base58 encoded address
    """
    tz_prefixes = {
        b'\x00\x00': b'tz1',
        b'\x00\x01': b'tz2',
        b'\x00\x02': b'tz3'
    }

    for bin_prefix, tz_prefix in tz_prefixes.items():
        if data.startswith(bin_prefix):
            return base58_encode(data[2:], tz_prefix).decode()

    if data.startswith(b'\x01') and data.endswith(b'\x00'):
        return base58_encode(data[1:-1], b'KT1').decode()
    else:
        return base58_encode(data[1:], tz_prefixes[b'\x00' + data[:1]]).decode()


def parse_contract(data: bytes):
    """ Decode contract (address + optional entrypoint) from bytes

    :param data: encoded contract
    :returns: base58 encoded address and entrypoint (if exists) separated by `%`
    """
    res = parse_address(data[:22])
    if len(data) > 22:
        res += f'%{data[22:].decode()}'
    return res


def forge_bool(value: bool) -> bytes:
    """ Encode boolean value into bytes.
    """
    return b'\xff' if value else b'\x00'


def forge_array(data, len_bytes=4) -> bytes:
    """ Encode array of bytes (prepend length).

    :param data: list of bytes
    :param len_bytes: number of bytes to store array length
    """
    return len(data).to_bytes(len_bytes, 'big') + data


def parse_array(data, len_bytes=4) -> tuple:
    """ Decode array of bytes.

    :param data: encoded array
    :param len_bytes: number of bytes to store array length
    :returns: Tuple[list of bytes, array length]
    """
    assert len(data) >= len_bytes, f'not enough bytes to parse array length, wanted {len_bytes}'
    length = int.from_bytes(data[:len_bytes], 'big')
    assert len(data) >= len_bytes + length, f'not enough bytes to parse array body, wanted {length}'
    return data[len_bytes:len_bytes+length], len_bytes+length


def forge_base58(value: str) -> bytes:
    """ Encode base58 string into bytes.

    :param value: base58 encoded value (with checksum)
    """
    return base58_decode(value.encode())


def forge_timestamp(value) -> int:
    """ Encode timestamp into bytes.

    :param value: unix timestamp in seconds (int)
    """
    assert isinstance(value, str)
    dt = datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
    return calendar.timegm(dt.utctimetuple())


def forge_contract(value) -> bytes:
    """ Encode a value of contract type (address + optional entrypoint) into bytes.

    :param value: 'tz12345' or 'tz12345%default'
    """
    parts = value.split('%')
    address, entrypoint = (parts[0], parts[1]) if len(parts) == 2 else (parts[0], 'default')
    res = forge_address(address)
    if entrypoint != 'default':
        res += entrypoint.encode()
    return res



from pytezos.rpc.node import RpcError


class MichelsonBadContractParameter(RpcError, error_id='michelson_v1.bad_contract_parameter'):
    """ Either no parameter was supplied to a contract with a non-unit parameter type, a non-unit parameter was passed
    to an account, or a parameter was supplied of the wrong type
    """


class MichelsonBadReturn(RpcError, error_id='michelson_v1.bad_return'):
    """ Unexpected stack at the end of a lambda or script
    """


class MichelsonRuntimeError(RpcError, error_id='michelson_v1'):
    """ Catch all michelson_v1 errors
    """


class TezArithmeticError(RpcError, error_id='tez'):
    """ Catch all tez errors
    """


class MichelsonScriptRejected(RpcError, error_id='script_rejected'):
    """ A FAILWITH instruction was reached
    """



from pprint import pformat

from pytezos.crypto import blake2b_32
from pytezos.operation.content import ContentMixin
from pytezos.operation.forge import forge_operation_group
from pytezos.operation.fees import calculate_fee, default_fee, default_gas_limit, default_storage_limit
from pytezos.operation.result import OperationResult
from pytezos.rpc.errors import RpcError
from pytezos.encoding import forge_base58, base58_encode
from pytezos.interop import Interop
from pytezos.tools.docstring import get_class_docstring

validation_passes = {
    'endorsement': 0,
    'proposal': 1,
    'ballot': 1,
    'seed_nonce_revelation': 2,
    'double_endorsement_evidence': 2,
    'double_baking_evidence': 2,
    'activate_account': 2,
    'reveal': 3,
    'transaction': 3,
    'origination': 3,
    'delegation': 3
}


class OperationGroup(Interop, ContentMixin):
    """ Operation group representation: contents (single or multiple), signature, other fields,
    and also useful helpers for filling with precise fees, signing, forging, and injecting.
    """

    def __init__(self, contents=None, protocol=None, chain_id=None, branch=None, signature=None, shell=None, key=None):
        super(OperationGroup, self).__init__(shell=shell, key=key)
        self.contents = contents or []
        self.protocol = protocol
        self.chain_id = chain_id
        self.branch = branch
        self.signature = signature

    def __repr__(self):
        res = [
            super(OperationGroup, self).__repr__(),
            '\nPayload',
            pformat(self.json_payload()),
            '\nHelpers',
            get_class_docstring(self.__class__)
        ]
        return '\n'.join(res)

    def _spawn(self, **kwargs):
        return OperationGroup(
            contents=kwargs.get('contents', self.contents.copy()),
            protocol=kwargs.get('protocol', self.protocol),
            chain_id=kwargs.get('chain_id', self.chain_id),
            branch=kwargs.get('branch', self.branch),
            signature=kwargs.get('signature', self.signature),
            shell=kwargs.get('shell', self.shell),
            key=kwargs.get('key', self.key)
        )

    def json_payload(self) -> dict:
        """ Get json payload used for the preapply.
        """
        return {
            'protocol': self.protocol,
            'branch': self.branch,
            'contents': self.contents,
            'signature': self.signature
        }

    def binary_payload(self) -> bytes:
        """ Get binary payload used for injection/hash calculation.
        """
        if not self.signature:
            raise ValueError('Not signed')

        return bytes.fromhex(self.forge()) + forge_base58(self.signature)

    def operation(self, content):
        """ Create new operation group with extra content added.

        :param content: Kind-specific operation body
        :rtype: OperationGroup
        """
        return self._spawn(contents=self.contents + [content])

    def fill(self):
        """ Try to fill all fields left unfilled, use approximate fees
        (not optimal, use `autofill` to simulate operation and get precise values).

        :rtype: OperationGroup
        """
        chain_id = self.chain_id or self.shell.chains.main.chain_id()
        branch = self.branch or self.shell.head.predecessor.hash()
        protocol = self.protocol or self.shell.head.header()['protocol']
        source = self.key.public_key_hash()
        counter = self.shell.contracts[source].count()

        replace_map = {
            'pkh': source,
            'source': source,
            'delegate': source,  # self registration
            'counter': lambda x: str(next(counter)),
            'secret': lambda x: self.key.activation_code,
            'period': lambda x: str(self.shell.head.voting_period()),
            'public_key': lambda x: self.key.public_key(),
            'manager_pubkey': source,  # I know, it hurts
            'fee': lambda x: str(default_fee(x)),
            'gas_limit': lambda x: str(default_gas_limit(x)),
            'storage_limit': lambda x: str(default_storage_limit(x)),
        }

        def fill_content(content):
            content = content.copy()
            for k, v in replace_map.items():
                if content.get(k) in ['', '0']:
                    content[k] = v(content) if callable(v) else v
            return content

        return self._spawn(
            contents=list(map(fill_content, self.contents)),
            protocol=protocol,
            chain_id=chain_id,
            branch=branch
        )

    def run(self):
        """ Simulate operation without signature checks.

        :returns: RPC response from `run_operation`
        """
        return self.shell.head.helpers.scripts.run_operation.post({
            'operation': {
                'branch': self.branch,
                'contents': self.contents,
                'signature': base58_encode(b'0' * 64, b'sig').decode()
            },
            'chain_id': self.chain_id
        })

    def forge(self, validate=True):
        """ Convert json representation of the operation group into bytes.

        :param validate: Forge remotely also and compare results, default is True
        :returns: Hex string
        """
        payload = {
            'branch': self.branch,
            'contents': self.contents
        }
        local_data = forge_operation_group(payload).hex()

        if validate:
            remote_data = self.shell.blocks[self.branch].helpers.forge.operations.post(payload)
            if local_data != remote_data:
                raise ValueError(f'Local forge result differs from remote one:\n\n{local_data}\n\n{remote_data}')

        return local_data

    def autofill(self, gas_reserve=100):
        """ Fill the gaps and then simulate the operation in order to calculate fee, gas/storage limits.

        :param gas_reserve: Add a safe reserve for gas limit (default is 100)
        :rtype: OperationGroup
        """
        opg = self.fill()
        opg_with_metadata = opg.run()
        if not OperationResult.is_applied(opg_with_metadata):
            raise RpcError.from_errors(OperationResult.errors(opg_with_metadata)) from None

        extra_size = (32 + 64) // len(opg.contents) + 1  # size of serialized branch and signature)

        def fill_content(content):
            if validation_passes[content['kind']] == 3:
                consumed_gas = OperationResult.consumed_gas(content) + gas_reserve
                paid_storage_size_diff = OperationResult.paid_storage_size_diff(content)
                burned = OperationResult.burned(content)
                fee = calculate_fee(content, consumed_gas, extra_size)
                content.update(
                    gas_limit=str(consumed_gas + gas_reserve),
                    storage_limit=str(paid_storage_size_diff + burned),
                    fee=str(fee)
                )

            content.pop('metadata')
            return content

        opg.contents = list(map(fill_content, opg_with_metadata['contents']))
        return opg

    def sign(self):
        """ Sign the operation group with the key specified by `using`.

        :rtype: OperationGroup
        """
        validation_pass = validation_passes[self.contents[0]['kind']]
        if any(map(lambda x: validation_passes[x['kind']] != validation_pass, self.contents)):
            raise ValueError('Mixed validation passes')

        if validation_pass == 0:
            chain_watermark = bytes.fromhex(self.shell.chains.main.watermark())
            watermark = b'\x02' + chain_watermark
        else:
            watermark = b'\x03'

        message = watermark + bytes.fromhex(self.forge())
        signature = self.key.sign(message=message, generic=True)

        return self._spawn(signature=signature)

    def hash(self) -> str:
        """ Calculate the Base58 encoded operation group hash.
        """
        hash_digest = blake2b_32(self.binary_payload()).digest()
        return base58_encode(hash_digest, b'o').decode()

    def preapply(self):
        """ Preapply signed operation group.

        :returns: RPC response from `preapply`
        """
        if not self.signature:
            raise ValueError('Not signed')

        return self.shell.head.helpers.preapply.operations.post(
            operations=[self.json_payload()])[0]

    def inject(self, _async=True, preapply=True, check_result=True, num_blocks_wait=2):
        """ Inject the signed operation group.

        :param _async: do not wait for operation inclusion (default is True)
        :param preapply: do a preapply before injection
        :param check_result: raise RpcError in case operation is refused
        :param num_blocks_wait: number of blocks to wait for injection
        :returns: operation group with metadata (raw RPC response)
        """
        if preapply:
            opg_with_metadata = self.preapply()
            if not OperationResult.is_applied(opg_with_metadata):
                raise RpcError.from_errors(OperationResult.errors(opg_with_metadata)) from None

        opg_hash = self.shell.injection.operation.post(
            operation=self.binary_payload(), _async=False)

        if _async:
            return {
                'chain_id': self.chain_id,
                'hash': opg_hash,
                **self.json_payload()
            }
        else:
            for i in range(num_blocks_wait):
                self.shell.wait_next_block()
                try:
                    pending_opg = self.shell.mempool.pending_operations[opg_hash]
                    if not OperationResult.is_applied(pending_opg):
                        raise RpcError.from_errors(OperationResult.errors(pending_opg)) from None
                    print(f'Still in mempool: {opg_hash}')
                except StopIteration:
                    res = self.shell.blocks[-(i + 1):].find_operation(opg_hash)
                    if check_result:
                        if not OperationResult.is_applied(res):
                            raise RpcError.from_errors(OperationResult.errors(res)) from None
                    return res

        raise TimeoutError(opg_hash)

    def result(self):
        """ Parse the preapply result.

        :rtype: OperationResult
        """
        return OperationResult.from_operation_group(self.preapply())



from pytezos.crypto import blake2b_32
from pytezos.encoding import forge_base58, forge_address, forge_public_key, forge_contract, forge_timestamp, \
    base58_encode
from pytezos.michelson.forge import forge_micheline, unforge_micheline
from pytezos.repl.parser import parse_prim_expr


def prepack_micheline(val_expr, type_expr):
    """ Recursively pack domain types into compact form,
    e.g. {"string": "tzabcd"} -> {"bytes": "0001deadbeef"}

    :param val_expr: value expression (Micheline expression)
    :param type_expr: type expression (Micheline expression)
    :returns: Micheline expression
    """
    def try_pack(val_node, type_node):
        type_prim, type_args = parse_prim_expr(type_node)
        is_string = isinstance(val_node, dict) and val_node.get('string')

        if type_prim in ['set', 'list']:
            val_node = [try_pack(i, type_args[0]) for i in val_node]
        elif type_prim in ['map', 'big_map']:
            val_node = [{'prim': 'Elt', 'args': [try_pack(elt['args'][i], type_args[i]) for i in [0, 1]]}
                        for elt in val_node]
        elif type_prim == 'pair':
            val_node['args'] = [try_pack(val_node['args'][i], type_args[i]) for i in [0, 1]]
        elif type_prim == 'option':
            if val_node['prim'] == 'Some':
                val_node['args'] = [try_pack(val_node['args'][0], type_args[0])]
        elif type_prim == 'or':
            type_idx = 0 if val_node['prim'] == 'Left' else 1
            val_node['args'] = [try_pack(val_node['args'][0], type_args[type_idx])]
        elif type_prim == 'lambda':
            pass  # TODO: PUSH, SELF, CONTRACT
        elif type_prim == 'chain_id' and is_string:
            return {'bytes': forge_base58(val_node['string']).hex()}
        elif type_prim == 'signature' and is_string:
            return {'bytes': forge_base58(val_node['string']).hex()}
        elif type_prim == 'key_hash' and is_string:
            return {'bytes': forge_address(val_node['string'], tz_only=True).hex()}
        elif type_prim == 'key' and is_string:
            return {'bytes': forge_public_key(val_node['string']).hex()}
        elif type_prim == 'address' and is_string:
            return {'bytes': forge_address(val_node['string']).hex()}
        elif type_prim == 'contract' and is_string:
            return {'bytes': forge_contract(val_node['string']).hex()}
        elif type_prim == 'timestamp' and is_string:
            return {'int': forge_timestamp(val_node['string'])}

        return val_node

    return try_pack(val_expr, type_expr)


def pack(val_expr, type_expr) -> bytes:
    """ Basically prepack (pack domain types) + forge (encode bytes). Resulting value is prefixed with 0x05.

    :param val_expr: value expression (Micheline expression)
    :param type_expr: type expression (Micheline expression)
    """
    data = prepack_micheline(val_expr, type_expr)
    return b'\x05' + forge_micheline(data)


def get_sub_expr(type_expr, bin_path='0'):
    assert len(bin_path) > 0, f'binary path should be at least `0`'
    node = type_expr
    for idx in bin_path[1:]:
        assert isinstance(node, dict), f'type expression contains dict nodes only'
        node = node['args'][int(idx)]
    return node


def get_key_hash(val_expr, type_expr, bin_path='') -> str:
    """ Get Big_map key hash from key and its type

    :param val_expr: key expression (Micheline expression)
    :param type_expr: type expression (can be key type or type of the whole storage)
    :param bin_path: binary path to the key (if storage type is passed to the prev argument)
    :returns: Base58 encoded key hash "expr..."
    """
    for idx in bin_path:
        assert isinstance(type_expr, dict), f'type expression contains dict nodes only'
        type_expr = type_expr['args'][int(idx)]

    data = blake2b_32(pack(val_expr, type_expr)).digest()
    return base58_encode(data, b'expr').decode()


def unpack(data: bytes, type_expr):
    """ Unpack bytes (currently without unpacking domain types, so it's unforging + cutting 0x05 prefix).

    :param data: Packed data
    :param type_expr: type of the packed data (currently unused)
    :returns: Micheline expression
    """
    assert data.startswith(b'\x05'), f'packed data should start with 05'
    parsed = unforge_micheline(data[1:])
    return parsed



import re
import functools
from collections import namedtuple
from typing import Tuple

COMPARE = dict(prim='COMPARE')
UNIT = dict(prim='UNIT')
FAILWITH = dict(prim='FAILWITH')
DUP = dict(prim='DUP')
SWAP = dict(prim='SWAP')
CAR = dict(prim='CAR')
CDR = dict(prim='CDR')
CAR__ = dict(prim='CAR', annots=['@%%'])
CDR__ = dict(prim='CDR', annots=['@%%'])
DROP = dict(prim='DROP')
FAIL = [[UNIT, FAILWITH]]

primitives = {
    'ABS', 'ADD', 'ADDRESS', 'AMOUNT', 'AND', 'BALANCE', 'BLAKE2B', 'CAR', 'CAST',
    'CDR', 'CHECK_SIGNATURE', 'COMPARE', 'CONCAT', 'CONS', 'CONTRACT',
    'CREATE_ACCOUNT', 'CREATE_CONTRACT', 'DIP', 'DROP', 'DUP', 'EDIV', 'EMPTY_MAP',
    'EMPTY_SET', 'EQ', 'EXEC', 'Elt', 'FAILWITH', 'False', 'GE', 'GET', 'GT',
    'HASH_KEY', 'IF', 'IF_CONS', 'IF_LEFT', 'IF_NONE', 'IMPLICIT_ACCOUNT', 'INT',
    'ISNAT', 'ITER', 'LAMBDA', 'LE', 'LEFT', 'LOOP', 'LOOP_LEFT', 'LSL', 'LSR',
    'LT', 'Left', 'MAP', 'MEM', 'MUL', 'NEG', 'NEQ', 'NIL', 'NONE', 'NOT', 'NOW',
    'None', 'OR', 'PACK', 'PAIR', 'PUSH', 'Pair', 'RENAME', 'RIGHT', 'Right',
    'SELF', 'SENDER', 'SET_DELEGATE', 'SHA256', 'SHA512', 'SIZE', 'SLICE', 'SOME',
    'SOURCE', 'STEPS_TO_QUOTA', 'SUB', 'SWAP', 'Some', 'TRANSFER_TOKENS', 'True',
    'UNIT', 'UNPACK', 'UPDATE', 'Unit', 'XOR', 'address', 'big_map', 'bool',
    'bytes', 'code', 'contract', 'int', 'key_hash', 'key', 'lambda', 'list', 'map',
    'mutez', 'nat', 'operation', 'option', 'or', 'pair', 'parameter', 'set',
    'signature', 'storage', 'string', 'timestamp', 'unit', 'DIG', 'DUG', 'EMPTY_BIG_MAP',
    'APPLY', 'chain_id', 'CHAIN_ID'
}
macros = []

PxrNode = namedtuple('PxrNode', ['depth', 'annots', 'args', 'is_root'])


def macro(regexp):
    def register_macro(func):
        macros.append((re.compile(regexp), func))
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return register_macro


def seq(instr=None) -> list:
    if instr is None:
        return []
    elif isinstance(instr, list):
        return instr
    else:
        return [instr]


def expand_macro(prim, annots, args, internal=False, extra=None):
    """ Expands Michelson macro.

    :param prim: macro name
    :param annots: annotations (optional)
    :param args: arguments (optional)
    :param internal: this function is called during another mocro expansion
    :param extra: list of ignored primitives
    :returns: Code sequence (Micheline expression)
    """
    assert isinstance(annots, list)
    assert isinstance(args, list)
    if prim in primitives or (isinstance(extra, list) and prim in extra):
        return expr(prim=prim, annots=annots, args=args)

    for regexp, handler in macros:
        groups = regexp.findall(prim)
        if groups:
            assert len(groups) == 1
            res = handler(groups[0], annots, args)
            return res if internal else seq(res)

    assert False, f'unknown primitive {prim}'


def get_field_annots(annots):
    return list(filter(lambda x: isinstance(x, str) and x[0] == '%', annots))


def get_var_annots(annots):
    return list(filter(lambda x: isinstance(x, str) and x[0] == '@', annots))


def skip_nones(array):
    return list(filter(lambda x: x is not None, array))


def expr(**kwargs) -> dict:
    return {k: v for k, v in kwargs.items() if v}


def dip_n(instr, depth=1):
    if depth <= 0:
        return instr
    elif depth == 1:
        return expr(prim='DIP', args=[seq(instr)])
    else:
        return expr(prim='DIP', args=[{'int': str(depth)}, seq(instr)])


@macro(r'^CMP(EQ|NEQ|LT|GT|LE|GE)$')
def expand_cmpx(prim, annots, args) -> list:
    assert not args
    return [COMPARE,
            expr(prim=prim, annots=annots)]


@macro(r'^IF(EQ|NEQ|LT|GT|LE|GE)$')
def expand_ifx(prim, annots, args) -> list:
    assert len(args) == 2
    return [expr(prim=prim, annots=annots),
            expr(prim='IF', args=args)]


@macro(r'^IFCMP(EQ|NEQ|LT|GT|LE|GE)$')
def expand_ifcmpx(prim, annots, args) -> list:
    assert len(args) == 2
    return [[COMPARE, expr(prim=prim, annots=annots)],
            expr(prim='IF', args=args)]


@macro(r'^FAIL$')
def expand_fail(prim, annots, args) -> list:
    assert not annots
    assert not args
    return [UNIT, FAILWITH]


@macro(r'^ASSERT$')
def expand_assert(prim, annots, args) -> dict:
    assert not annots
    assert not args
    return expr(prim='IF', args=[[], FAIL])


@macro(r'^ASSERT_(EQ|NEQ|LT|LE|GT|GE)$')
def expand_assert_x(prim, annots, args) -> list:
    assert not args
    assert not annots  # TODO: ask why
    return expand_ifx(prim, annots=[], args=[[], FAIL])


@macro(r'^ASSERT_CMP(EQ|NEQ|LT|LE|GT|GE)$')
def expand_assert_cmpx(prim, annots, args) -> list:
    assert not args
    assert not annots  # TODO: ask why
    return expand_ifcmpx(prim, annots=[], args=[[], FAIL])


@macro(r'^ASSERT_NONE$')
def expand_assert_none(prim, annots, args) -> dict:
    assert not annots
    assert not args
    return expr(prim='IF_NONE', args=[[], FAIL])


@macro('^ASSERT_SOME$')
def expand_assert_some(prim, annots, args) -> dict:
    assert not args
    return expr(prim='IF_NONE',
                args=[FAIL, [expr(prim='RENAME', annots=annots)]])


@macro('^ASSERT_LEFT$')
def expand_assert_left(prim, annots, args) -> dict:
    assert not args
    return expr(prim='IF_LEFT',
                args=[[expr(prim='RENAME', annots=annots)], FAIL])


@macro('^ASSERT_RIGHT$')
def expand_assert_right(prim, annots, args) -> dict:
    assert not args
    return expr(prim='IF_LEFT',
                args=[FAIL, [expr(prim='RENAME', annots=annots)]])


@macro(r'^D(II+)P$')
def expand_dixp(prim, annots, args) -> dict:
    assert not annots
    assert len(args) == 1
    return dip_n(args, depth=len(prim))


@macro(r'^D(UU+)P$')
def expand_duxp(prim, annots, args) -> list:
    assert not args
    depth = len(prim)
    dup = expr(prim='DUP', annots=annots)
    if depth == 1:
        return [dup]
    elif depth == 2:
        return [dip_n(dup), SWAP]
    else:
        return [dip_n(dup, depth - 1),
                expr(prim='DIG', args=[{'int': str(depth)}])]


def build_pxr_tree(pxr_macro, pxr_annots) -> PxrNode:
    def parse(prim, annots, depth=0, is_root=False):
        letter, prim = prim[0], prim[1:]
        if letter == 'P':
            dip_depth = depth
            left, l_annot, prim, annots, depth = parse(prim, annots, depth)
            right, r_annot, prim, annots, depth = parse(prim, annots, depth)
            return PxrNode(dip_depth, [l_annot, r_annot], [left, right], is_root), None, prim, annots, depth
        else:
            annot, annots = (annots[0], annots[1:]) if annots else (None, [])
            return letter, annot, prim, annots, depth + 1
    root, _, _, _, _ = parse(pxr_macro, pxr_annots, is_root=True)
    return root


def traverse_pxr_tree(prim, annots, produce):
    res = []

    def walk(node):
        if isinstance(node, PxrNode):
            res.insert(0, dip_n(produce(node), depth=node.depth))
            _ = list(map(walk, node.args))

    walk(build_pxr_tree(prim, annots))
    return res


@macro(r'^P[PAI]{3,}R$')
def expand_pxr(prim, annots, args) -> list:
    def produce(node: PxrNode):
        pair_annots = [node.annots[0] or '%', node.annots[1]] if any(node.annots) else []
        if node.is_root:
            pair_annots.extend(get_var_annots(annots))
        return expr(prim='PAIR', annots=skip_nones(pair_annots))

    assert not args
    return traverse_pxr_tree(prim, get_field_annots(annots), produce)


@macro(r'^UNPAIR$')
def expand_unpair(prim, annots, args) -> list:
    assert not args
    assert len(annots) in {0, 2}
    car_annots, cdr_annots = (seq(annots[0]), seq(annots[1])) if annots else ([], [])
    return [DUP,
            expr(prim='CAR', annots=car_annots),
            dip_n(expr(prim='CDR', annots=cdr_annots))]


@macro(r'^UN(P[PAI]{3,}R)$')
def expand_unpxr(prim, annots, args) -> list:
    def produce(node: PxrNode):
        return expand_unpair(prim=None, annots=node.annots, args=[])

    assert not args
    return list(reversed(traverse_pxr_tree(prim, annots, produce)))


def expand_cxr(prim, annots) -> list:
    return seq(expand_macro(prim=f'C{prim}R', annots=annots, args=[], internal=True))


@macro(r'^CA([AD]+)R$')
def expand_caxr(prim, annots, args) -> list:
    assert not args
    return [CAR, *expand_cxr(prim, annots)]


@macro(r'^CD([AD]+)R$')
def expand_cdxr(prim, annots, args) -> list:
    assert not args
    return [CDR, *expand_cxr(prim, annots)]


@macro(r'^IF_SOME$')
def expand_if_some(prim, annots, args) -> dict:
    assert not annots
    assert len(args) == 2
    return expr(prim='IF_NONE', args=list(reversed(args)))


@macro(r'^SET_CAR$')
def expand_set_car(prim, annots, args) -> list:
    assert not args
    if annots:
        assert len(annots) == 1
        access_check = [DUP,
                        expr(prim='CAR', annots=annots),
                        DROP]
    else:
        access_check, annots = [], ['%']
    return [*access_check,
            CDR__,
            SWAP,
            expr(prim='PAIR', annots=[annots[0], '%@'])]


@macro(r'^SET_CDR$')
def expand_set_cdr(prim, annots, args) -> list:
    assert not args
    if annots:
        assert len(annots) == 1
        access_check = [DUP,
                        expr(prim='CDR', annots=annots),
                        DROP]
    else:
        access_check, annots = [], ['%']
    return [*access_check,
            CAR__,
            expr(prim='PAIR', annots=['%@', annots[0]])]


def expand_set_cxr(prim, annots):
    set_cxr = expand_macro(prim=f'SET_C{prim}R', annots=get_field_annots(annots), args=[], internal=True)
    pair = expr(prim='PAIR', annots=['%@', '%@'] + get_var_annots(annots))
    return set_cxr, pair


@macro(r'^SET_CA([AD]+)R$')
def expand_set_caxr(prim, annots, args) -> list:
    assert not args
    set_cxr, pair = expand_set_cxr(prim, annots)
    return [DUP,
            dip_n([CAR__, set_cxr]),
            CDR__,
            SWAP,
            pair]


@macro(r'^SET_CD([AD]+)R$')
def expand_set_cdxr(prim, annots, args) -> list:
    assert not args
    set_cxr, pair = expand_set_cxr(prim, annots)
    return [DUP,
            dip_n([CDR__, set_cxr]),
            CAR__,
            pair]


def get_map_cxr_annots(annots) -> Tuple[str, list]:
    field_annots = get_field_annots(annots)
    if field_annots:
        assert len(field_annots) == 1
        return field_annots[0], [f'@{field_annots[0][1:]}']
    else:
        return '%', []


@macro(r'^MAP_CAR$')
def expand_map_car(prim, annots, args) -> list:
    car_annot, var_annots = get_map_cxr_annots(annots)
    return [DUP,
            CDR__,
            dip_n([expr(prim='CAR', annots=var_annots), *args]),
            SWAP,
            expr(prim='PAIR', annots=[car_annot, '%@'])]


@macro(r'^MAP_CDR$')
def expand_map_cdr(prim, annots, args) -> list:
    cdr_annot, var_annots = get_map_cxr_annots(annots)
    return [DUP,
            expr(prim='CDR', annots=var_annots),
            *args,
            SWAP,
            CAR__,
            expr(prim='PAIR', annots=['%@', cdr_annot])]


def expand_map_cxr(prim, annots, args):
    set_cxr = expand_macro(prim=f'MAP_C{prim}R', annots=get_field_annots(annots), args=args, internal=True)
    pair = expr(prim='PAIR', annots=['%@', '%@'] + get_var_annots(annots))
    return set_cxr, pair


@macro(r'^MAP_CA([AD]+)R$')
def expand_map_caxr(prim, annots, args) -> list:
    map_cxr, pair = expand_map_cxr(prim, annots, args)
    return [DUP,
            dip_n([CAR__, map_cxr]),
            CDR__,
            SWAP,
            pair]


@macro(r'^MAP_CD([AD]+)R$')
def expand_map_cdxr(prim, annots, args) -> list:
    map_cxr, pair = expand_map_cxr(prim, annots, args)
    return [DUP,
            dip_n([CDR__, map_cxr]),
            CAR__,
            pair]



"""
Welcome to PyTezos!

To start playing with the Tezos blockchain you need to get a PyTezosClient instance.
Just type:

>>> from pytezos import pytezos
>>> pytezos

And follow the interactive documentation.
"""

from pytezos.rpc import RpcProvider, localhost, mainnet, babylonnet, zeronet
from pytezos.rpc.errors import *
from pytezos.crypto import Key
from pytezos.proto import Proto
from pytezos.michelson.micheline import Unit
from pytezos.michelson.contract import Contract
from pytezos.michelson.formatter import format_timestamp
from pytezos.client import PyTezosClient
from pytezos.operation.group import OperationGroup
from pytezos.michelson.interface import ContractInterface
from pytezos.standards.non_fungible_token import NonFungibleTokenImpl

pytezos = PyTezosClient()



from functools import lru_cache

from pytezos.michelson.contract import Contract, ContractParameter, ContractStorage
from pytezos.michelson.micheline import michelson_to_micheline

parameter_tz = """
parameter 
    (or (or (nat %burn :token_id) (pair %mint (address %owner) (nat %token_id)))
        (pair %transfer (address %destination) (nat %token_id)))
"""
storage_tz = "storage (map nat address)"


class NonFungibleTokenImpl(Contract):
    """ This is an example of how to force contract interface (in case it's not annotated)
    """

    @property
    @lru_cache(maxsize=None)
    def parameter(self) -> ContractParameter:
        return ContractParameter(michelson_to_micheline(parameter_tz))

    @property
    @lru_cache(maxsize=None)
    def storage(self) -> ContractStorage:
        return ContractStorage(michelson_to_micheline(storage_tz))



import yaml
from copy import deepcopy
from pprint import pformat

from pytezos.michelson.grammar import MichelsonParser, MichelsonParserError
from pytezos.michelson.converter import michelson_to_micheline, micheline_to_michelson
from pytezos.repl.control import do_interpret, MichelsonRuntimeError
from pytezos.repl.helpers import *
from pytezos.repl.arithmetic import *
from pytezos.repl.structures import *
from pytezos.repl.blockchain import *


def get_content(obj: dict):
    content = obj.get('_content')
    if content:
        return format_content(content)
    return {}


def format_stack_item(item: StackItem):
    row = {}
    if isinstance(item, Operation):
        row['value'] = yaml.dump(get_content(item.val_expr)).rstrip('\n')
    elif isinstance(item, List) and item.val_type() == Operation:
        row['value'] = yaml.dump([get_content(x) for x in item.val_expr]).rstrip('\n')
    else:
        row['value'] = micheline_to_michelson(item.val_expr)

    row['type'] = micheline_to_michelson(item.type_expr)
    if item.name is not None:
        row['name'] = f'@{item.name}'

    return row


def format_diff(diff: dict):
    if diff['action'] == 'alloc':
        return {'big_map': diff['big_map'],
                'action': diff['action'],
                'key': micheline_to_michelson(diff['key_type']),
                'value': micheline_to_michelson(diff['value_type'])}
    elif diff['action'] == 'update':
        return {'big_map': diff['big_map'],
                'action': diff['action'],
                'key': micheline_to_michelson(diff['key']),
                'value': micheline_to_michelson(diff['value']) if diff.get('value') else 'null'}
    elif diff['action'] == 'copy':
        return {'destination_big_map': diff['big_map'],
                'action': diff['action'],
                'value': diff['source_big_map']}
    elif diff['action'] == 'remove':
        return {'big_map': diff['big_map'],
                'action': diff['action']}
    else:
        assert False, diff['action']


def format_content(content):
    if content['kind'] == 'transaction':
        return {'kind': content['kind'],
                'target': content['destination'],
                'amount': content['amount'],
                'entrypoint': content['parameters']['entrypoint'],
                'parameters': micheline_to_michelson(content['parameters']['value'])}
    elif content['kind'] == 'origination':
        res = {'kind': content['kind'],
               'target': content['originated_contract'],
               'amount': content['balance'],
               'storage': micheline_to_michelson(content['script']['storage']),
               'code': micheline_to_michelson(content['script']['code'])}
        if content.get('delegate'):
            res['delegate'] = content['delegate']
        return res
    elif content['kind'] == 'delegation':
        return {'kind': content['kind'],
                'target': content['delegate']}
    else:
        assert False, content['kind']


def format_stdout(items):
    newline = True
    depth = 0
    res = []

    def break_line():
        nonlocal newline
        nonlocal res
        if not newline:
            res.append('\n')
            newline = True

    for item in items:
        if item['action'] == 'begin':
            break_line()
            if item['prim'] not in ['parameter', 'storage', 'code', 'STORAGE', 'DUMP']:
                res.extend(['  ' * depth, item['prim'], ':'])
                depth += 1
                newline = False
        elif item['action'] == 'end':
            depth -= 1
            break_line()
        elif item['action'] in ['message', 'event']:
            res.append('  ' * depth if newline else ' ')
            res.extend([item['text'], ';'])
            newline = False
        else:
            assert False, item['action']

    res = ''.join(res).strip('\n')
    if '\n' not in res and all(map(lambda x: x['action'] != 'message', items)):
        return ''

    return res


def format_result(result):
    if result is None:
        return None
    kind = result['kind']
    if kind == 'message':
        return result
    elif kind == 'big_map_diff':
        return {'value': list(map(format_diff, result['big_map_diff'])), **result}
    elif kind == 'code':
        return {'value': micheline_to_michelson(result['code']), **result}
    elif kind == 'stack':
        return {'value': list(map(format_stack_item, result['stack'])), **result}
    elif kind == 'output':
        operations = [format_content(x.get('_content')) for x in result['operations'].val_expr]
        storage = [format_stack_item(result['storage'])]
        big_map_diff = list(map(format_diff, result['big_map_diff']))
        return {'value': (operations, storage, big_map_diff), **result}
    else:
        assert False, kind


def format_stderr(error):
    ename = type(error).__name__
    if isinstance(error, MichelsonRuntimeError):
        evalue, traceback = error.message, 'at ' + ' -> '.join(error.trace)
    elif isinstance(error, MichelsonParserError):
        evalue, traceback = error.message, f'at line {error.line}, pos {error.pos}'
    else:
        evalue, traceback = pformat(error.args, compact=True), ''
    return {'name': ename,
            'value': evalue,
            'trace': traceback}


class Interpreter:
    """ Michelson interpreter reimplemented in Python.
    Based on the following reference: https://michelson.nomadic-labs.com/
    """

    def __init__(self, debug=True):
        self.ctx = Context()
        self.parser = MichelsonParser(extra_primitives=helpers_prim)
        self.debug = debug

    def execute(self, code):
        """ Execute Michelson instructions (note that stack is not cleared after execution).

        :param code: Michelson source (any valid Michelson expression or special helpers)
        :returns: {"success": True|False, "stdout": "", "stderr": {}, "result": {"value": "", ...}}
        """
        int_res = {'success': False}

        try:
            code_expr = michelson_to_micheline(code, parser=self.parser)
        except MichelsonParserError as e:
            if self.debug:
                raise e
            int_res['stderr'] = format_stderr(e)
            return int_res

        backup = deepcopy(self.ctx)

        try:
            res = do_interpret(self.ctx, code_expr)
            if res is None and self.ctx.pushed:
                res = {'kind': 'stack', 'stack': self.ctx.dump(count=1)}

            int_res['result'] = format_result(res)
            int_res['stdout'] = format_stdout(self.ctx.stdout)
            int_res['success'] = True
            self.ctx.reset()
        except MichelsonRuntimeError as e:
            int_res['stderr'] = format_stderr(e)
            int_res['stdout'] = format_stdout(self.ctx.stdout)
            self.ctx = backup

            if self.debug:
                if int_res.get('stdout'):
                    print(int_res['stdout'])
                raise e

        if self.debug:
            if int_res.get('stdout'):
                print(int_res['stdout'])
            if int_res.get('result'):
                print('RESULT: ' + pformat(int_res['result']))

        return int_res



import hashlib
import binascii
import json
from os.path import expanduser, join, abspath
from getpass import getpass
from pyblake2 import blake2b
from mnemonic import Mnemonic

from pytezos.encoding import scrub_input, base58_decode, base58_encode
from pytezos.tools.docstring import InlineDocstring, get_class_docstring


class CryptoExtraFallback:

    def __getattr__(self, item):
        raise NotImplementedError(
            "Please, install packages libsodium-dev, libsecp256k1-dev, and libgmp-dev, "
            "and Python libraries pysodium, secp256k1, and fastecdsa")

    def __call__(self, *args, **kwargs):
        self.__getattr__('throw')


try:
    import pysodium
    import secp256k1
    import fastecdsa.ecdsa
    import fastecdsa.keys
    import fastecdsa.curve
    import fastecdsa.encoding.sec1
    from fastecdsa.encoding.util import bytes_to_int
except ImportError as e:
    pysodium = CryptoExtraFallback()
    secp256k1 = CryptoExtraFallback()
    fastecdsa = CryptoExtraFallback()
    bytes_to_int = CryptoExtraFallback()
    __crypto__ = False
else:
    __crypto__ = True


def is_installed():
    return __crypto__


def blake2b_32(v=b''):
    """ Get a BLAKE2B hash of bytes
    """
    return blake2b(scrub_input(v), digest_size=32)


def validate_mnemonic(mnemonic, language='english'):
    m = Mnemonic(language)
    mnemonic = m.normalize_string(mnemonic).split(' ')
    if len(mnemonic) not in [12, 15, 18, 21, 24]:
        raise ValueError(
            'Number of words must be one of the following: [12, 15, 18, 21, 24], but it is not (%d).' % len(mnemonic))

    idx = map(lambda x: bin(m.wordlist.index(x))[2:].zfill(11), mnemonic)
    b = ''.join(idx)
    l = len(b)
    d = b[:l // 33 * 32]
    h = b[-l // 33:]
    nd = binascii.unhexlify(hex(int(d, 2))[2:].rstrip('L').zfill(l // 33 * 8))
    nh = bin(int(hashlib.sha256(nd).hexdigest(), 16))[2:].zfill(256)[:l // 33]
    if h != nh:
        raise ValueError('Failed checksum.')


class Key(metaclass=InlineDocstring):
    """ Represents a public or secret key for Tezos. Ed25519, Secp256k1 and P256
    are supported.
    """
    def __init__(self, public_point, secret_exponent=None, curve=b'ed', activation_code=None):
        self.public_point = public_point
        self.secret_exponent = secret_exponent
        self.curve = curve
        self.activation_code = activation_code

    def __repr__(self):
        res = [
            super(Key, self).__repr__(),
            f'\nPublic key hash',
            self.public_key_hash(),
            '\nHelpers',
            get_class_docstring(self.__class__)
        ]
        return '\n'.join(res)

    @property
    def is_secret(self):
        return self.secret_exponent is not None

    @classmethod
    def from_secret_exponent(cls, secret_exponent: bytes, curve=b'ed', activation_code=None):
        """ Creates a key object from a secret exponent.

        :param secret_exponent: secret exponent or seed
        :param curve: b'sp' for Secp251k1, b'p2' for P256/Secp256r1, b'ed' for Ed25519 (default)
        :param activation_code: secret for initializing account balance
        """
        # Ed25519
        if curve == b'ed':
            # Dealing with secret exponent or seed?
            if len(secret_exponent) == 64:
                public_point = pysodium.crypto_sign_sk_to_pk(sk=secret_exponent)
            else:
                public_point, secret_exponent = pysodium.crypto_sign_seed_keypair(seed=secret_exponent)
        # Secp256k1
        elif curve == b'sp':
            sk = secp256k1.PrivateKey(secret_exponent)
            public_point = sk.pubkey.serialize()
        # P256
        elif curve == b'p2':
            pk = fastecdsa.keys.get_public_key(bytes_to_int(secret_exponent), curve=fastecdsa.curve.P256)
            public_point = fastecdsa.encoding.sec1.SEC1Encoder.encode_public_key(pk)
        else:
            assert False

        return cls(public_point, secret_exponent, curve=curve, activation_code=activation_code)

    @classmethod
    def from_public_point(cls, public_point: bytes, curve=b'ed'):
        """ Creates a key object from a public elliptic point.

        :param public_point: elliptic point in the compressed format (see https://tezos.stackexchange.com/a/623/309)
        :param curve: b'sp' for secp251k1, b'p2' for P256/secp256r1, b'ed' for Ed25519 (default)
        """
        return cls(public_point, curve=curve)

    @classmethod
    def from_encoded_key(cls, key, passphrase=''):
        """ Creates a key object from a base58 encoded key.

        :param key: a public or secret key in base58 encoding
        :param passphrase: the passphrase used if the key provided is an encrypted private key
        """
        key = scrub_input(key)

        curve = key[:2]  # "sp", "p2" "ed"
        if curve not in [b'sp', b'p2', b'ed']:
            raise ValueError("Invalid prefix for a key encoding.")
        if not len(key) in [54, 55, 88, 98]:
            raise ValueError("Invalid length for a key encoding.")

        encrypted = (key[2:3] == b'e')
        public_or_secret = key[3:5] if encrypted else key[2:4]
        if public_or_secret not in [b'pk', b'sk']:
            raise Exception("Invalid prefix for a key encoding.")

        key = base58_decode(key)
        is_secret = (public_or_secret == b'sk')
        if not is_secret:
            return cls.from_public_point(key, curve)

        if encrypted:
            if not passphrase:
                raise ValueError("Encrypted key provided without a passphrase.")

            salt, encrypted_sk = key[:8], key[8:]
            encryption_key = hashlib.pbkdf2_hmac(
                hash_name="sha512",
                password=scrub_input(passphrase),
                salt=salt,
                iterations=32768,
                dklen=32
            )
            key = pysodium.crypto_secretbox_open(
                c=encrypted_sk, nonce=b'\000' * 24, k=encryption_key)
            del passphrase

        return cls.from_secret_exponent(key, curve)

    @classmethod
    def generate(cls, passphrase='', curve=b'ed', strength=128, language='english', export=True):
        """ Generates new key.

        :param passphrase: optional password
        :param curve: b'sp' for secp251k1, b'p2' for P256/secp256r1, b'ed' for Ed25519 (default)
        :param strength: mnemonic strength, default is 128
        :param language: mnemonic language, default is english
        :param export: export as json file in the current folder, default is True
        :rtype: Key
        """
        mnemonic = Mnemonic(language).generate(strength)
        key = cls.from_mnemonic(mnemonic, passphrase, curve=curve)

        if export:
            pkh = key.public_key_hash()
            data = {
                'mnemonic': mnemonic.split(),
                'pkh': pkh,
                'password': passphrase
            }
            with open(abspath(f'./{pkh}.json'), 'w+') as f:
                f.write(json.dumps(data))

        return key

    @classmethod
    def from_mnemonic(cls, mnemonic, passphrase='', email='', validate=True, curve=b'ed', activation_code=None):
        """ Creates a key object from a bip39 mnemonic.

        :param mnemonic: a 15 word bip39 english mnemonic
        :param passphrase: a mnemonic password or a fundraiser key
        :param email: email used if a fundraiser key is passed
        :param validate: whether to check mnemonic or not
        :param curve: b'sp' for secp251k1, b'p2' for P256/secp256r1, b'ed' for Ed25519 (default)
        :param activation_code: secret for initializing account balance
        :rtype: Key
        """
        if isinstance(mnemonic, list):
            mnemonic = ' '.join(mnemonic)

        if validate:
            validate_mnemonic(mnemonic)

        seed = Mnemonic.to_seed(mnemonic, passphrase=email + passphrase)

        if curve == b'ed':
            _, secret_exponent = pysodium.crypto_sign_seed_keypair(seed=seed[:32])
        elif curve == b'sp':
            secret_exponent = seed[:32]
        elif curve == b'p2':
            secret_exponent = seed[:32]
        else:
            assert False

        return cls.from_secret_exponent(secret_exponent, curve=curve, activation_code=activation_code)

    @classmethod
    def from_faucet(cls, path):
        """ Import key from a faucet file: https://faucet.tzalpha.net/

        :param path: path to the json file
        :rtype: Key
        """
        with open(expanduser(path), 'r') as f:
            data = json.loads(f.read())

        key = cls.from_mnemonic(
            mnemonic=data['mnemonic'],
            passphrase=data.get('password', ''),
            email=data.get('email', ''),
            activation_code=data['secret']
        )
        if key.public_key_hash() != data['pkh']:
            raise ValueError('Failed to import')

        return key

    @classmethod
    def from_alias(cls, alias, passphrase='', tezos_client_dir='~/.tezos-client'):
        """ Import secret key from tezos-client keychain.

        :param alias: key alias
        :param passphrase: if key is encrypted (optional)
        :param tezos_client_dir: path to the tezos client directory (default is `~/.tezos-client`)
        :rtype: Key
        """
        path = expanduser(join(tezos_client_dir, 'secret_keys'))
        with open(path, 'r') as f:
            data = json.loads(f.read())

        value = next(x['value'] for x in data if x['name'] == alias)
        prefix, sk = value.split(':', maxsplit=1)

        if prefix == 'encrypted':
            if not passphrase:
                passphrase = getpass(f'Please, enter passphrase for `{alias}`:\n')
            key = cls.from_encoded_key(sk, passphrase=passphrase)
            del passphrase
        else:
            key = cls.from_encoded_key(sk)

        return key

    def public_key(self):
        """ Creates base58 encoded public key representation.

        :returns: the public key associated with the private key
        """
        return base58_encode(self.public_point, self.curve + b'pk').decode()

    def secret_key(self, passphrase=None, ed25519_seed=True):
        """ Creates base58 encoded private key representation.

        :param passphrase: encryption phrase for the private key
        :param ed25519_seed: encode seed rather than full key for ed25519 curve (True by default)
        :returns: the secret key associated with this key, if available
        """
        if not self.secret_exponent:
            raise ValueError("Secret key not known.")

        if self.curve == b'ed' and ed25519_seed:
            key = pysodium.crypto_sign_sk_to_seed(self.secret_exponent)
        else:
            key = self.secret_exponent

        if passphrase:
            if not ed25519_seed:
                raise NotImplementedError

            salt = pysodium.randombytes(8)
            encryption_key = hashlib.pbkdf2_hmac(
                hash_name="sha512",
                password=scrub_input(passphrase),
                salt=salt,
                iterations=32768,
                dklen=32
            )
            encrypted_sk = pysodium.crypto_secretbox(
                msg=key, nonce=b'\000' * 24, k=encryption_key)
            key = salt + encrypted_sk  # we have to combine salt and encrypted key in order to decrypt later
            prefix = self.curve + b'esk'
        else:
            prefix = self.curve + b'sk'

        return base58_encode(key, prefix).decode()

    def public_key_hash(self):
        """ Creates base58 encoded public key hash for this key.

        :returns: the public key hash for this key
        """
        pkh = blake2b(data=self.public_point, digest_size=20).digest()
        prefix = {b'ed': b'tz1', b'sp': b'tz2', b'p2': b'tz3'}[self.curve]
        return base58_encode(pkh, prefix).decode()

    def sign(self, message, generic=False):
        """ Sign a raw sequence of bytes.

        :param message: sequence of bytes, raw format or hexadecimal notation
        :param generic: do not specify elliptic curve if set to True
        :returns: signature in base58 encoding
        """
        message = scrub_input(message)

        if not self.secret_exponent:
            raise ValueError("Cannot sign without a secret key.")

        # Ed25519
        if self.curve == b"ed":
            digest = pysodium.crypto_generichash(message)
            signature = pysodium.crypto_sign_detached(digest, self.secret_exponent)
        # Secp256k1
        elif self.curve == b"sp":
            pk = secp256k1.PrivateKey(self.secret_exponent)
            signature = pk.ecdsa_serialize_compact(
                pk.ecdsa_sign(message, digest=blake2b_32))
        # P256
        elif self.curve == b"p2":
            r, s = fastecdsa.ecdsa.sign(msg=message, d=bytes_to_int(self.secret_exponent), hashfunc=blake2b_32)
            signature = r.to_bytes(32, 'big') + s.to_bytes(32, 'big')
        else:
            assert False

        if generic:
            prefix = b'sig'
        else:
            prefix = self.curve + b'sig'

        return base58_encode(signature, prefix).decode()

    def verify(self, signature, message):
        """ Verify signature, raise exception if it is not valid.

        :param message: sequance of bytes, raw format or hexadecimal notation
        :param signature: a signature in base58 encoding
        :raises: ValueError if signature is not valid
        """
        signature = scrub_input(signature)
        message = scrub_input(message)

        if not self.public_point:
            raise ValueError("Cannot verify without a public key")

        if signature[:3] != b'sig':  # not generic
            if self.curve != signature[:2]:  # "sp", "p2" "ed"
                raise ValueError("Signature and public key curves mismatch.")

        signature = base58_decode(signature)

        # Ed25519
        if self.curve == b"ed":
            digest = pysodium.crypto_generichash(message)
            try:
                pysodium.crypto_sign_verify_detached(signature, digest, self.public_point)
            except ValueError:
                raise ValueError('Signature is invalid.')
        # Secp256k1
        elif self.curve == b"sp":
            pk = secp256k1.PublicKey(self.public_point, raw=True)
            sig = pk.ecdsa_deserialize_compact(signature)
            if not pk.ecdsa_verify(message, sig, digest=blake2b_32):
                raise ValueError('Signature is invalid.')
        # P256
        elif self.curve == b"p2":
            pk = fastecdsa.encoding.sec1.SEC1Encoder.decode_public_key(self.public_point, curve=fastecdsa.curve.P256)
            r, s = bytes_to_int(signature[:32]), bytes_to_int(signature[32:])
            if not fastecdsa.ecdsa.verify(sig=(r, s), msg=message, Q=pk, hashfunc=blake2b_32):
                raise ValueError('Signature is invalid.')
        else:
            assert False



from pytezos.encoding import forge_array, parse_array

prim_tags = {
    'parameter': b'\x00',
    'storage': b'\x01',
    'code': b'\x02',
    'False': b'\x03',
    'Elt': b'\x04',
    'Left': b'\x05',
    'None': b'\x06',
    'Pair': b'\x07',
    'Right': b'\x08',
    'Some': b'\x09',
    'True': b'\x0A',
    'Unit': b'\x0B',
    'PACK': b'\x0C',
    'UNPACK': b'\x0D',
    'BLAKE2B': b'\x0E',
    'SHA256': b'\x0F',
    'SHA512': b'\x10',
    'ABS': b'\x11',
    'ADD': b'\x12',
    'AMOUNT': b'\x13',
    'AND': b'\x14',
    'BALANCE': b'\x15',
    'CAR': b'\x16',
    'CDR': b'\x17',
    'CHECK_SIGNATURE': b'\x18',
    'COMPARE': b'\x19',
    'CONCAT': b'\x1A',
    'CONS': b'\x1B',
    '__CREATE_ACCOUNT__': b'\x1C',  # DEPRECATED
    'CREATE_CONTRACT': b'\x1D',
    'IMPLICIT_ACCOUNT': b'\x1E',
    'DIP': b'\x1F',
    'DROP': b'\x20',
    'DUP': b'\x21',
    'EDIV': b'\x22',
    'EMPTY_MAP': b'\x23',
    'EMPTY_SET': b'\x24',
    'EQ': b'\x25',
    'EXEC': b'\x26',
    'FAILWITH': b'\x27',
    'GE': b'\x28',
    'GET': b'\x29',
    'GT': b'\x2A',
    'HASH_KEY': b'\x2B',
    'IF': b'\x2C',
    'IF_CONS': b'\x2D',
    'IF_LEFT': b'\x2E',
    'IF_NONE': b'\x2F',
    'INT': b'\x30',
    'LAMBDA': b'\x31',
    'LE': b'\x32',
    'LEFT': b'\x33',
    'LOOP': b'\x34',
    'LSL': b'\x35',
    'LSR': b'\x36',
    'LT': b'\x37',
    'MAP': b'\x38',
    'MEM': b'\x39',
    'MUL': b'\x3A',
    'NEG': b'\x3B',
    'NEQ': b'\x3C',
    'NIL': b'\x3D',
    'NONE': b'\x3E',
    'NOT': b'\x3F',
    'NOW': b'\x40',
    'OR': b'\x41',
    'PAIR': b'\x42',
    'PUSH': b'\x43',
    'RIGHT': b'\x44',
    'SIZE': b'\x45',
    'SOME': b'\x46',
    'SOURCE': b'\x47',
    'SENDER': b'\x48',
    'SELF': b'\x49',
    'STEPS_TO_QUOTA': b'\x4A',
    'SUB': b'\x4B',
    'SWAP': b'\x4C',
    'TRANSFER_TOKENS': b'\x4D',
    'SET_DELEGATE': b'\x4E',
    'UNIT': b'\x4F',
    'UPDATE': b'\x50',
    'XOR': b'\x51',
    'ITER': b'\x52',
    'LOOP_LEFT': b'\x53',
    'ADDRESS': b'\x54',
    'CONTRACT': b'\x55',
    'ISNAT': b'\x56',
    'CAST': b'\x57',
    'RENAME': b'\x58',
    'bool': b'\x59',
    'contract': b'\x5A',
    'int': b'\x5B',
    'key': b'\x5C',
    'key_hash': b'\x5D',
    'lambda': b'\x5E',
    'list': b'\x5F',
    'map': b'\x60',
    'big_map': b'\x61',
    'nat': b'\x62',
    'option': b'\x63',
    'or': b'\x64',
    'pair': b'\x65',
    'set': b'\x66',
    'signature': b'\x67',
    'string': b'\x68',
    'bytes': b'\x69',
    'mutez': b'\x6A',
    'timestamp': b'\x6B',
    'unit': b'\x6C',
    'operation': b'\x6D',
    'address': b'\x6E',
    'SLICE': b'\x6F',
    'DIG': b'\x70',
    'DUG': b'\x71',
    'EMPTY_BIG_MAP': b'\x72',
    'APPLY': b'\x73',
    'chain_id': b'\x74',
    'CHAIN_ID': b'\x75'
}
prim_int = {v[0]: k for k, v in prim_tags.items()}

len_tags = [
    {
        False: b'\x03',
        True: b'\x04'
    },
    {
        False: b'\x05',
        True: b'\x06'
    },
    {
        False: b'\x07',
        True: b'\x08'
    },
    {
        False: b'\x09',
        True: b'\x09'
    }
]

reserved_entries = {
    'default': b'\x00',
    'root': b'\x01',
    'do': b'\x02',
    'set_delegate': b'\x03',
    'remove_delegate': b'\x04'
}


def forge_int(value: int) -> bytes:
    """
    Encode signed unbounded integer into the byte form.

    :param value: signed unbounded integer
    """
    res = bytearray()
    i = abs(value)

    res.append((i & 0b00111111) | (0b11000000 if value < 0 else 0b10000000))
    i >>= 6

    while i != 0:
        res.append((i & 0b01111111) | 0b10000000)
        i >>= 7

    res[-1] &= 0b01111111
    return bytes(res)


def parse_int(data: bytes) -> (int, int):
    """ Decode signed unbounded integer from bytes.

    :param data: Encoded integer
    :returns: tuple(parsed integer, length in bytes)
    """
    value = 0
    length = 1

    while data[length - 1] & 0b10000000 != 0:
        length += 1

    for i in range(length - 1, 0, -1):
        value <<= 7
        value |= data[i] & 0b01111111

    value <<= 6
    value |= data[0] & 0b00111111

    if (data[0] & 0b01000000) != 0:
        value = -value

    return value, length


def forge_entrypoint(entrypoint) -> bytes:
    """ Encode Michelson contract entrypoint into the byte form.

    :param entrypoint: string
    """
    if entrypoint in reserved_entries:
        return reserved_entries[entrypoint]
    else:
        return b'\xff' + forge_array(entrypoint.encode(), len_bytes=1)


def forge_micheline(data) -> bytes:
    """ Encode a Micheline expression into the byte form.

    :param data: Micheline expression
    """
    res = []

    if isinstance(data, list):
        res.append(b'\x02')
        res.append(forge_array(b''.join(map(forge_micheline, data))))

    elif isinstance(data, dict):
        if data.get('prim'):
            args_len = len(data.get('args', []))
            annots_len = len(data.get('annots', []))

            res.append(len_tags[args_len][annots_len > 0])
            res.append(prim_tags[data['prim']])

            if args_len > 0:
                args = b''.join(map(forge_micheline, data['args']))
                if args_len < 3:
                    res.append(args)
                else:
                    res.append(forge_array(args))

            if annots_len > 0:
                res.append(forge_array(' '.join(data['annots']).encode()))
            elif args_len == 3:
                res.append(b'\x00' * 4)

        elif data.get('bytes') is not None:
            res.append(b'\x0A')
            res.append(forge_array(bytes.fromhex(data['bytes'])))

        elif data.get('int') is not None:
            res.append(b'\x00')
            res.append(forge_int(int(data['int'])))

        elif data.get('string') is not None:
            res.append(b'\x01')
            res.append(forge_array(data['string'].encode()))
        else:
            assert False, data
    else:
        assert False, data

    return b''.join(res)


def forge_script(script) -> bytes:
    """ Encode an origination script into the byte form.

    :param script: {"code": "$Micheline_expression", "storage": "$Micheline_expression"}
    """
    code = forge_micheline(script['code'])
    storage = forge_micheline(script['storage'])
    return forge_array(code) + forge_array(storage)


def unforge_micheline(data: bytes):
    """ Parse Micheline expression from its encoded form (but do not UNPACK).

    :param data: Encoded Micheline expression
    :returns: Object
    """
    ptr = 0

    def parse_list():
        nonlocal ptr
        _, offset = parse_array(data[ptr:])
        end, res = ptr + offset, []
        ptr += 4
        while ptr < end:
            res.append(parse())
        assert ptr == end, f'out of array boundaries'
        return res

    def parse_prim_expr(args_len=0, annots=False):
        nonlocal ptr
        prim_tag = data[ptr]
        ptr += 1
        expr = {'prim': prim_int[prim_tag]}

        if 0 < args_len < 3:
            expr['args'] = [parse() for _ in range(args_len)]
        elif args_len == 3:
            expr['args'] = parse_list()
        else:
            assert args_len == 0, f'unexpected args len {args_len}'

        if annots:
            value, offset = parse_array(data[ptr:])
            ptr += offset
            if len(value) > 0:
                expr['annots'] = value.decode().split(' ')

        return expr

    def parse():
        nonlocal ptr
        tag = data[ptr]
        ptr += 1
        if tag == 0:
            value, offset = parse_int(data[ptr:])
            ptr += offset
            return {'int': str(value)}
        elif tag == 1:
            value, offset = parse_array(data[ptr:])
            ptr += offset
            return {'string': value.decode()}
        elif tag == 2:
            return parse_list()
        elif tag == 3:
            return parse_prim_expr(args_len=0, annots=False)
        elif tag == 4:
            return parse_prim_expr(args_len=0, annots=True)
        elif tag == 5:
            return parse_prim_expr(args_len=1, annots=False)
        elif tag == 6:
            return parse_prim_expr(args_len=1, annots=True)
        elif tag == 7:
            return parse_prim_expr(args_len=2, annots=False)
        elif tag == 8:
            return parse_prim_expr(args_len=2, annots=True)
        elif tag == 9:
            return parse_prim_expr(args_len=3, annots=True)
        elif tag == 10:
            value, offset = parse_array(data[ptr:])
            ptr += offset
            return {'bytes': value.hex()}
        else:
            assert False, f'unkonwn tag {tag} at position {ptr}'

    res = parse()
    assert ptr == len(data), f'have not reach EOS (pos {ptr}/{len(data)})'
    return res



from pytezos.michelson.pack import pack
from pytezos.repl.control import instruction
from pytezos.repl.context import Context
from pytezos.repl.types import assert_stack_type, Option, Pair, String, Bytes, List, BigMap, Map, Set, Or, Bool, Nat, \
    Unit, StackItem, dispatch_type_map
from pytezos.michelson.pack import unpack


@instruction(['CAR', 'CDR'])
def do_car(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Pair)
    idx = {'CAR': 0, 'CDR': 1}
    res = top.get_element(idx[prim])
    ctx.push(res, annots=annots)


@instruction('CONCAT')
def do_concat(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, [String, Bytes, List])
    if type(top) in [String, Bytes]:
        second = ctx.pop1()
        val_type = dispatch_type_map(top, second, {
            (String, String): str,
            (Bytes, Bytes): bytes
        })
        res = type(top)(val_type(top) + val_type(second))
    elif type(top) == List:
        res_type = top.val_type()
        val_type, sep = {
            String: (str, ''),
            Bytes: (bytes, b'')
        }[res_type]
        res = res_type(sep.join(map(val_type, top)))
    else:
        assert False
    ctx.push(res, annots=annots)


@instruction('CONS')
def do_cons(ctx: Context, prim, args, annots):
    val, container = ctx.pop2()
    assert_stack_type(container, List)
    res = container.prepend(val)
    ctx.push(res, annots=annots)


@instruction('EMPTY_BIG_MAP', args_len=2)
def do_empty_big_map(ctx: Context, prim, args, annots):
    res = ctx.big_maps.empty(k_type_expr=args[0], v_type_expr=args[1])
    ctx.push(res, annots=annots)


@instruction('EMPTY_MAP', args_len=2)
def do_empty_map(ctx: Context, prim, args, annots):
    res = Map.empty(k_type_expr=args[0], v_type_expr=args[1])
    ctx.push(res, annots=annots)


@instruction('EMPTY_SET', args_len=1)
def do_empty_set(ctx: Context, prim, args, annots):
    res = Set.empty(k_type_expr=args[0])
    ctx.push(res, annots=annots)


@instruction('GET')
def do_get(ctx: Context, prim, args, annots):
    key, container = ctx.pop2()
    assert_stack_type(container, [Map, BigMap])

    if type(container) == Map:
        val = container.find(key)
    else:
        val = ctx.big_maps.find(container, key)

    if val is not None:
        res = Option.some(val)
    else:
        res = Option.none(container.val_type_expr())

    ctx.push(res, annots=annots)


@instruction(['LEFT', 'RIGHT'], args_len=1)
def do_left(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    if prim == 'LEFT':
        res = Or.left(r_type_expr=args[0], item=top)
    else:
        res = Or.right(l_type_expr=args[0], item=top)
    ctx.push(res, annots=annots)


@instruction('MEM')
def do_mem(ctx: Context, prim, args, annots):
    key, container = ctx.pop2()
    assert_stack_type(container, [Set, Map, BigMap])
    if type(container) == BigMap:
        res = Bool(ctx.big_maps.contains(container, key))
    else:
        res = Bool(key in container)
    ctx.push(res, annots=annots)


@instruction('NIL', args_len=1)
def do_nil(ctx: Context, prim, args, annots):
    nil = List.empty(args[0])
    ctx.push(nil, annots=annots)


@instruction('NONE', args_len=1)
def do_none(ctx: Context, prim, args, annots):
    none = Option.none(args[0])
    ctx.push(none, annots=annots)


@instruction('PACK')
def do_pack(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    res = Bytes(pack(top.val_expr, top.type_expr))
    ctx.push(res, annots=annots)


@instruction('PAIR')
def do_pair(ctx: Context, prim, args, annots):
    left, right = ctx.pop2()
    res = Pair.new(left, right)
    ctx.push(res, annots=annots)


@instruction('SIZE')
def do_size(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, [String, Bytes, List, Set, Map])
    res = Nat(len(top))
    ctx.push(res, annots=annots)


@instruction('SLICE')
def do_slice(ctx: Context, prim, args, annots):
    offset, length, s = ctx.pop3()
    assert_stack_type(s, [String, Bytes])
    offset, length = int(offset), int(length)
    if len(s) > 0 and offset + length <= len(s):
        res = Option.some(s[offset:offset+length])
    else:
        res = Option.none(type(s)().type_expr)
    ctx.push(res, annots=annots)


@instruction('SOME')
def do_some(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    res = Option.some(top)
    ctx.push(res, annots=annots)


@instruction('UNIT')
def do_unit(ctx: Context, prim, args, annots):
    ctx.push(Unit(), annots=annots)


@instruction('UNPACK', args_len=1)
def do_unpack(ctx: Context, prim, args, annots):
    top = ctx.pop1()
    assert_stack_type(top, Bytes)
    try:
        val_expr = unpack(data=bytes(top), type_expr=args[0])
        item = StackItem.parse(val_expr=val_expr, type_expr=args[0])
        res = Option.some(item)
    except Exception as e:
        ctx.print(f'failed: {e}')
        res = Option.none(args[0])
    ctx.push(res, annots=annots)


@instruction('UPDATE')
def do_update(ctx: Context, prim, args, annots):
    key, val, container = ctx.pop3()
    assert_stack_type(container, [Set, Map, BigMap])

    if type(container) == Set:
        assert_stack_type(val, Bool)
        if val:
            res = container.add(key)
        else:
            res = container.remove(key)
    else:
        assert_stack_type(val, Option)
        if val.is_none():
            res = container.remove(key)
        else:
            res = container.update(key, val.get_some())

    ctx.push(res, annots=annots)



from pytezos.rpc.query import RpcQuery


class BakingRightsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/baking_rights'):

    def __call__(self, level=None, cycle=None, delegate=None, max_priority=None, _all=None):
        """ Retrieves the list of delegates allowed to bake a block.
        By default, it gives the best baking priorities for bakers that have at least one opportunity
        below the 64th priority for the next block.

        :param level: Specify the (valid) level in the past or future at which the baking rights have to be returned.
        :param cycle: Specify the (valid) levels in the past or future at which the baking rights have to be returned.
        :param delegate: Restrict the results to the given delegates.
        :param max_priority:
        :param _all: If parameter `all` is set, all the baking opportunities for each baker at each level are returned,\
         instead of just the first one.
        :returns: Returns the list of baking slots. Also returns the minimal timestamps that correspond to these slots.\
         The timestamps are omitted for levels in the past, and are only estimates for levels later that the next block,\
         based on the hypothesis that all predecessor blocks were baked at the first priority.
        """
        return self._get(params={
            'level': level,
            'cycle': cycle,
            'delegate': delegate,
            'max_priority': max_priority,
            'all': _all
        })


class ForgeOperationsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/forge/operations'):

    def post(self, operation):
        """ Get raw bytes encoding of an unsigned operation

        :param operation: Json input
            {
                "branch": <block hash>,
                "contents": [{ <operation_content> }]
            }
        :returns: Hex string
        """
        return self._post(json=operation)


class ForgeProtocolDataQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/forge/protocol_data'):

    def post(self, protocol_data):
        """ Forge the protocol-specific part of a block header.

        :param protocol_data: Json input
            {
                "priority": <integer>,
                "nonce_hash": <Base58 nonce hash>,  // optional
                "proof_of_work_nonce": <hex encoded>  // optional
            }
        :returns: { "protocol_data": <hex encoded> }
        """
        return self._post(json=protocol_data)


class ForgeBlockHeaderQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/forge_block_header'):

    def post(self, block_header):
        """ Forge block header.

        :param block_header: Json input
            {
                "level": <integer>,
                "proto": <integer>,
                "predecessor": <block hash>,
                "timestamp": <timestamp>,
                "validation_pass": <integer>,
                "operations_hash": <Operation_list_list_hash>,
                "fitness": <integer (hex)>,
                "context": <Context_hash>,
                "protocol_data": <hex encoded>
            }
        :returns: { "block": <hex encoded> }
        """
        return self._post(json=block_header)


class ParseBlockQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/parse/block'):

    def post(self, block_header):
        """ Retrieves protocol-specific part of a block header and signature.

        :param block_header: Json input
            {
                "level": <integer>,
                "proto": <integer>,
                "predecessor": <block hash>,
                "timestamp": <timestamp>,
                "validation_pass": <integer>,
                "operations_hash": <Operation_list_list_hash>,
                "fitness": <integer (hex)>,
                "context": <Context_hash>,
                "protocol_data": <hex encoded>
            }
        :returns: Json object
        """
        return self._post(json=block_header)


class ParseOperationsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/parse/operations'):

    def post(self, operations):
        """ Extracts contents and signatures from the forged and optionally signed operations (bulk).

        :param operations: Json input
            {
                "operations": [ {
                    "branch": <block_hash>,
                    "data": <hex encoded>
                } ... ],
                "check_signature": <boolean>  // optional
            }
        :returns: Json object
        """
        return self._post(json=operations)


class PreapplyBlockQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/preapply/block'):

    def post(self, block, sort=None, timestamp=None):
        """ Simulate the validation of a block that would contain the given operations
        and return the resulting fitness and context hash.

        :param block: Json input
            {
                "protocol_data": {
                    "protocol": "Pt24m4xiPbLDhVgVfABUjirbmda3yohdN82Sp9FeuAXJ4eV9otd",
                    "priority": <integer>,
                    "proof_of_work_nonce": <hex encoded>
                    "seed_nonce_hash": <Base58 encoded>,  // optional
                    "signature": <Base58 encoded>
                },
                "operations": [ [ {
                    "protocol": "Pt24m4xiPbLDhVgVfABUjirbmda3yohdN82Sp9FeuAXJ4eV9otd",
                    "branch": <block_hash>,
                    "contents": [ {} ... ],  // kind-specific
                    "signature": <Base58 encoded>
                }... ] ... ]
            }
        :param sort: by what?
        :param timestamp: timestamp
        :returns: Json object
        """
        return self._post(
            json=block,
            params=dict(sort=sort, timestamp=timestamp)
        )


class PreapplyOperationsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/preapply/operations'):

    def post(self, operations):
        """ Simulate the validation of operation(s).

        :param operations: Json input
            [{
                "protocol": "Pt24m4xiPbLDhVgVfABUjirbmda3yohdN82Sp9FeuAXJ4eV9otd",
                "branch": <block_hash>,
                "contents": [ {} ... ],  // kind-specific
                "signature": <Base58 encoded>
            }]
        :returns: Json array (operations with metadata)
        """
        return self._post(json=operations)


class ScriptsEntrypoint(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/entrypoint'):

    def post(self, script, entrypoint=None):
        """ Return the type of the given entrypoint.

        :param script: Micheline expression
        :param entrypoint: Name of the entrypoint (leave None if there's only one)
        """
        payload = {'script': script}
        if entrypoint:
            payload['entrypoint'] = entrypoint

        return self._post(payload)


class ScriptsEntrypoints(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/entrypoints'):

    def post(self, script):
        """ Return the list of entrypoints of the given script.

        :param script: Micheline expression
        """
        return self._post({'script': script})


class ScriptsPackDataQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/pack_data'):

    def post(self, expression):
        """ Computes the serialized version of some data expression using the same algorithm as script instruction PACK.

        :param expression: Json input
            {
                "data": <michelson expression>,
                "type": <michelson expression>,
                "gas": <bignum>  // optional
            }
        :returns: Packed data (hex encoded)
        """
        return self._post(json=expression)


class ScriptsRunCodeQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/run_code'):

    def post(self, invocation):
        """ Run a piece of code in the current context.

        :param invocation: Json input
            {
                "script": <michelson expression>,
                "storage": <michelson expression>,
                "input": <michelson expression>,
                "amount": <mutez>,
                "source": <account address>,  // optional
                "payer": <account address>,  // optional
                "gas": <bignum>  // optional
            }
        :returns: Resulting storage, spawned operations, and bigmap diff
        """
        return self._post(json=invocation)


class ScriptsRunOperationQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/run_operation'):

    def post(self, operation):
        """ Run an operation without signature checks.

        :param operation: Json input
            {
                "branch": <block_hash>,
                "contents": [ {} ... ],  // kind-specific
                "signature": <Base58 encoded>
            }
        :returns: Json object (operation with metadata)
        """
        return self._post(json=operation)


class ScriptsTraceCodeQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/trace_code'):

    def post(self, invocation):
        """ Run a piece of code in the current context, keeping a trace.

        :param invocation: Json input
            {
                "script": <michelson expression>,
                "storage": <michelson expression>,
                "input": <michelson expression>,
                "amount": <mutez>,
                "source": <account address>,  // optional
                "payer": <account address>,  // optional
                "gas": <bignum>  // optional
            }
        :returns: Resulting storage, spawned operations, bigmap diff, and trace
        """
        return self._post(json=invocation)


class ScriptsTypecheckCodeQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/typecheck_code'):

    def post(self, expression):
        """ Check that some data expression is well formed and of a given type in the current context.

        :param expression: Json input
            {
                "data": <michelson expression>,
                "type": <michelson expression>,
                "gas": <bignum>  // optional
            }
        :returns: Json object
        """
        return self._post(json=expression)



from os.path import basename

from pytezos.michelson.micheline import Schema, is_optional

core_types = ['string', 'int', 'bool']
domain_types = {
    'nat': 'int  /* Natural number */',
    'unit': 'Unit || None /* Void */',
    'bytes': 'string  /* Hex string */ ||\n\tbytes  /* Python byte string */',
    'timestamp': 'int  /* Unix time in seconds */ ||\n\tstring  /* Formatted datetime `%Y-%m-%dT%H:%M:%SZ` */',
    'mutez': 'int  /* Amount in `utz` (10^-6) */ ||\n\tDecimal  /* Amount in `tz` */',
    'contract': 'string  /* Base58 encoded `KT` address */',
    'address': 'string  /* Base58 encoded `tz` or `KT` address */',
    'key': 'string  /* Base58 encoded public key */',
    'key_hash': 'string  /* Base58 encoded public key hash */',
    'signature': 'string  /* Base58 encoded signature */',
    'lambda': 'string  /* Michelson source code */'
}


def generate_docstring(schema: Schema, title, root='0') -> str:
    """ Generate Michelson type (of arbitrary complexity) documentation in a more readable form.

    :param schema: parameter/storage schema
    :param title: documentation title
    :param root: binary path to the root element (default is '0')
    :returns: formatted docstring
    """
    docstring = list()
    known_types = set()

    def get_node(bin_path):
        return schema.metadata[bin_path]

    def get_name(bin_path):
        name = schema.bin_names.get(bin_path)
        if not name:
            name = next((basename(k) for k, v in schema.json_to_bin.items() if v == bin_path), bin_path)
        return name

    def get_type(bin_path):
        default = get_name(bin_path[:-1]) + '_item'
        return schema.metadata[bin_path].get('typename', default)

    def get_comment(bin_path):
        node = schema.metadata[bin_path]
        return node.get('typename', node.get('fieldname'))

    def decode_node(bin_path, is_element=False, is_entry=False):
        node = get_node(bin_path)
        bin_type = schema.bin_types[bin_path]

        def get_struct_name():
            if bin_path == root:
                struct_name = title
            elif is_element:
                struct_name = get_type(bin_path)
            else:
                struct_name = get_name(bin_path)
            return f'${struct_name}'

        if bin_type == 'router':
            entries = {get_name(x): decode_node(x, is_entry=True) for x in node['args']}
            doc = ' || \n\t'.join(map(lambda x: '{ ' + f'"{x[0]}": {x[1]}' + ' }', entries.items()))
            res = get_struct_name()
            docstring.insert(0, f'{res}:\n\t{doc}\n')
            return res

        elif bin_type == 'enum':
            res = ' || '.join(map(lambda x: f'"{get_name(x)}"', node['args']))

        elif bin_type == 'namedtuple':
            items = map(lambda x: (get_name(x), decode_node(x)), node['args'])
            lines = map(lambda x: f'  "{x[0]}": {x[1]}', items)
            doc = '\t{\n\t' + ',\n\t'.join(lines) + '\n\t}'
            res = get_struct_name()
            docstring.insert(0, f'{res}:\n{doc}\n')
            return res

        elif bin_type == 'tuple':
            values = map(decode_node, node['args'])
            res = f'[ {" , ".join(values)} ]'

        elif bin_type in {'keypair', 'pair'}:
            values = map(decode_node, node['args'])
            res = f'( {" , ".join(values)} )'

        elif bin_type in {'set', 'list'}:
            value = decode_node(node['args'][0], is_element=True)
            res = f'[ {value} , ... ]'

        elif bin_type in {'map', 'big_map'}:
            item = (decode_node(node['args'][0]), decode_node(node['args'][1], is_element=True))
            res = '{ ' + f'{item[0]} : {item[1]} , ...' + ' }'
            if bin_type == 'big_map':
                res += '  /* big_map */'

        else:
            res = node['prim']
            if res not in core_types:
                res = f'${res}'

            if is_entry:
                comment = get_comment(bin_path)
                if comment:
                    res = f'{res}  /* {comment} */'

            if node['prim'] in ['contract', 'lambda']:
                parameter = schema.metadata[bin_path]['parameter']
                res = f'{res} ({parameter})'

            if is_optional(schema, bin_path):
                res = f'None || {res}'

            if node['prim'] not in core_types:
                if bin_path == root:
                    res = domain_types[node["prim"]]
                else:
                    assert node['prim'] in domain_types, f'not a domain type {node["prim"]}'
                    known_types.add(node['prim'])

        if bin_path == root:
            docstring.insert(0, f'${title}:\n\t{res}\n')

        return res

    decode_node(root)

    for prim in known_types:
        docstring.append(f'${prim}:\n\t{domain_types[prim]}\n')

    return '\n'.join(docstring)
